
//
// This source file is part of appleseed.
// Visit http://appleseedhq.net/ for additional information and resources.
//
// This software is released under the MIT license.
//
// Copyright (c) 2010-2013 Francois Beaune, Jupiter Jazz Limited
// Copyright (c) 2014 Francois Beaune, The appleseedhq Organization
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
//

// Interface header.
#include "permutation.h"

// appleseed.foundation headers.
#include "foundation/math/primes.h"

namespace foundation
{

//
// Permutation generators implementation.
//

void faure_qmc_permutation(
    const size_t    size,
    size_t          perm[])
{
    assert(size >= 2);

    if (size == 2)
    {
        // Identity permutation.
        perm[0] = 0;
        perm[1] = 1;
    }
    else if (size & 1)
    {
        // Recursively build Faure permutation for odd size.
        faure_qmc_permutation(size - 1, perm);
        const size_t k = (size - 1) >> 1;
        for (size_t i = 0; i < size - 1; ++i)
        {
            if (perm[i] >= k)
                ++perm[i];
        }

        // Insert k in the middle of the permutation.
        for (size_t i = size - 1; i >= k + 1; --i)
            perm[i] = perm[i - 1];
        perm[k] = k;
    }
    else
    {
        // Recursively build Faure permutation for even size.
        const size_t k = size >> 1;
        faure_qmc_permutation(k, perm);
        for (size_t i = 0; i < k; ++i)
        {
            perm[i] <<= 1;
            perm[i + k] = perm[i] + 1;
        }
    }
}


//
// Precomputed Faure permutations.
//

const size_t FaurePermutationValues[24133] =
{
    // Size 2
      0,   1, 
    // Size 3
      0,   1,   2, 
    // Size 5
      0,   3,   2,   1,   4, 
    // Size 7
      0,   2,   5,   3,   1,   4,   6, 
    // Size 11
      0,   7,   4,   2,   9,   5,   1,   8,   6,   3,
     10, 
    // Size 13
      0,   4,   9,   2,   7,  11,   6,   1,   5,  10,
      3,   8,  12, 
    // Size 17
      0,   9,   4,  13,   2,  11,   6,  15,   8,   1,
     10,   5,  14,   3,  12,   7,  16, 
    // Size 19
      0,  11,   4,  15,   8,   2,  13,   6,  17,   9,
      1,  12,   5,  16,  10,   3,  14,   7,  18, 
    // Size 23
      0,  15,   8,   4,  19,  10,   2,  17,  13,   6,
     21,  11,   1,  16,   9,   5,  20,  12,   3,  18,
     14,   7,  22, 
    // Size 29
      0,   8,  21,  12,   4,  17,  25,   2,  10,  23,
     15,   6,  19,  27,  14,   1,   9,  22,  13,   5,
     18,  26,   3,  11,  24,  16,   7,  20,  28, 
    // Size 31
      0,   8,  23,  12,   4,  19,  27,  14,   2,  10,
     25,  17,   6,  21,  29,  15,   1,   9,  24,  13,
      5,  20,  28,  16,   3,  11,  26,  18,   7,  22,
     30, 
    // Size 37
      0,  21,   8,  29,  16,   4,  25,  12,  33,   2,
     23,  10,  31,  19,   6,  27,  14,  35,  18,   1,
     22,   9,  30,  17,   5,  26,  13,  34,   3,  24,
     11,  32,  20,   7,  28,  15,  36, 
    // Size 41
      0,  25,  16,   8,  33,   4,  29,  21,  12,  37,
      2,  27,  18,  10,  35,   6,  31,  23,  14,  39,
     20,   1,  26,  17,   9,  34,   5,  30,  22,  13,
     38,   3,  28,  19,  11,  36,   7,  32,  24,  15,
     40, 
    // Size 43
      0,  27,  16,   8,  35,   4,  31,  23,  12,  39,
     20,   2,  29,  18,  10,  37,   6,  33,  25,  14,
     41,  21,   1,  28,  17,   9,  36,   5,  32,  24,
     13,  40,  22,   3,  30,  19,  11,  38,   7,  34,
     26,  15,  42, 
    // Size 47
      0,  31,  16,   8,  39,  20,   4,  35,  27,  12,
     43,  22,   2,  33,  18,  10,  41,  25,   6,  37,
     29,  14,  45,  23,   1,  32,  17,   9,  40,  21,
      5,  36,  28,  13,  44,  24,   3,  34,  19,  11,
     42,  26,   7,  38,  30,  15,  46, 
    // Size 53
      0,  16,  37,   8,  29,  45,  24,   4,  20,  41,
     12,  33,  49,   2,  18,  39,  10,  31,  47,  27,
      6,  22,  43,  14,  35,  51,  26,   1,  17,  38,
      9,  30,  46,  25,   5,  21,  42,  13,  34,  50,
      3,  19,  40,  11,  32,  48,  28,   7,  23,  44,
     15,  36,  52, 
    // Size 59
      0,  16,  43,  24,   8,  35,  51,   4,  20,  47,
     31,  12,  39,  55,  28,   2,  18,  45,  26,  10,
     37,  53,   6,  22,  49,  33,  14,  41,  57,  29,
      1,  17,  44,  25,   9,  36,  52,   5,  21,  48,
     32,  13,  40,  56,  30,   3,  19,  46,  27,  11,
     38,  54,   7,  23,  50,  34,  15,  42,  58, 
    // Size 61
      0,  16,  45,  24,   8,  37,  53,  28,   4,  20,
     49,  33,  12,  41,  57,   2,  18,  47,  26,  10,
     39,  55,  31,   6,  22,  51,  35,  14,  43,  59,
     30,   1,  17,  46,  25,   9,  38,  54,  29,   5,
     21,  50,  34,  13,  42,  58,   3,  19,  48,  27,
     11,  40,  56,  32,   7,  23,  52,  36,  15,  44,
     60, 
    // Size 67
      0,  35,  16,  51,   8,  43,  24,  59,   4,  39,
     20,  55,  12,  47,  28,  63,  32,   2,  37,  18,
     53,  10,  45,  26,  61,   6,  41,  22,  57,  14,
     49,  30,  65,  33,   1,  36,  17,  52,   9,  44,
     25,  60,   5,  40,  21,  56,  13,  48,  29,  64,
     34,   3,  38,  19,  54,  11,  46,  27,  62,   7,
     42,  23,  58,  15,  50,  31,  66, 
    // Size 71
      0,  39,  16,  55,   8,  47,  24,  63,  32,   4,
     43,  20,  59,  12,  51,  28,  67,  34,   2,  41,
     18,  57,  10,  49,  26,  65,  37,   6,  45,  22,
     61,  14,  53,  30,  69,  35,   1,  40,  17,  56,
      9,  48,  25,  64,  33,   5,  44,  21,  60,  13,
     52,  29,  68,  36,   3,  42,  19,  58,  11,  50,
     27,  66,  38,   7,  46,  23,  62,  15,  54,  31,
     70, 
    // Size 73
      0,  41,  16,  57,  32,   8,  49,  24,  65,   4,
     45,  20,  61,  37,  12,  53,  28,  69,   2,  43,
     18,  59,  34,  10,  51,  26,  67,   6,  47,  22,
     63,  39,  14,  55,  30,  71,  36,   1,  42,  17,
     58,  33,   9,  50,  25,  66,   5,  46,  21,  62,
     38,  13,  54,  29,  70,   3,  44,  19,  60,  35,
     11,  52,  27,  68,   7,  48,  23,  64,  40,  15,
     56,  31,  72, 
    // Size 79
      0,  47,  16,  63,  32,   8,  55,  24,  71,  36,
      4,  51,  20,  67,  43,  12,  59,  28,  75,  38,
      2,  49,  18,  65,  34,  10,  57,  26,  73,  41,
      6,  53,  22,  69,  45,  14,  61,  30,  77,  39,
      1,  48,  17,  64,  33,   9,  56,  25,  72,  37,
      5,  52,  21,  68,  44,  13,  60,  29,  76,  40,
      3,  50,  19,  66,  35,  11,  58,  27,  74,  42,
      7,  54,  23,  70,  46,  15,  62,  31,  78, 
    // Size 83
      0,  51,  32,  16,  67,   8,  59,  43,  24,  75,
      4,  55,  36,  20,  71,  12,  63,  47,  28,  79,
     40,   2,  53,  34,  18,  69,  10,  61,  45,  26,
     77,   6,  57,  38,  22,  73,  14,  65,  49,  30,
     81,  41,   1,  52,  33,  17,  68,   9,  60,  44,
     25,  76,   5,  56,  37,  21,  72,  13,  64,  48,
     29,  80,  42,   3,  54,  35,  19,  70,  11,  62,
     46,  27,  78,   7,  58,  39,  23,  74,  15,  66,
     50,  31,  82, 
    // Size 89
      0,  57,  32,  16,  73,  40,   8,  65,  49,  24,
     81,   4,  61,  36,  20,  77,  45,  12,  69,  53,
     28,  85,   2,  59,  34,  18,  75,  42,  10,  67,
     51,  26,  83,   6,  63,  38,  22,  79,  47,  14,
     71,  55,  30,  87,  44,   1,  58,  33,  17,  74,
     41,   9,  66,  50,  25,  82,   5,  62,  37,  21,
     78,  46,  13,  70,  54,  29,  86,   3,  60,  35,
     19,  76,  43,  11,  68,  52,  27,  84,   7,  64,
     39,  23,  80,  48,  15,  72,  56,  31,  88, 
    // Size 97
      0,  32,  65,  16,  49,  81,   8,  40,  73,  24,
     57,  89,   4,  36,  69,  20,  53,  85,  12,  44,
     77,  28,  61,  93,   2,  34,  67,  18,  51,  83,
     10,  42,  75,  26,  59,  91,   6,  38,  71,  22,
     55,  87,  14,  46,  79,  30,  63,  95,  48,   1,
     33,  66,  17,  50,  82,   9,  41,  74,  25,  58,
     90,   5,  37,  70,  21,  54,  86,  13,  45,  78,
     29,  62,  94,   3,  35,  68,  19,  52,  84,  11,
     43,  76,  27,  60,  92,   7,  39,  72,  23,  56,
     88,  15,  47,  80,  31,  64,  96, 
    // Size 101
      0,  32,  69,  16,  53,  85,   8,  40,  77,  24,
     61,  93,  48,   4,  36,  73,  20,  57,  89,  12,
     44,  81,  28,  65,  97,   2,  34,  71,  18,  55,
     87,  10,  42,  79,  26,  63,  95,  51,   6,  38,
     75,  22,  59,  91,  14,  46,  83,  30,  67,  99,
     50,   1,  33,  70,  17,  54,  86,   9,  41,  78,
     25,  62,  94,  49,   5,  37,  74,  21,  58,  90,
     13,  45,  82,  29,  66,  98,   3,  35,  72,  19,
     56,  88,  11,  43,  80,  27,  64,  96,  52,   7,
     39,  76,  23,  60,  92,  15,  47,  84,  31,  68,
    100, 
    // Size 103
      0,  32,  71,  16,  55,  87,   8,  40,  79,  24,
     63,  95,  48,   4,  36,  75,  20,  59,  91,  12,
     44,  83,  28,  67,  99,  50,   2,  34,  73,  18,
     57,  89,  10,  42,  81,  26,  65,  97,  53,   6,
     38,  77,  22,  61,  93,  14,  46,  85,  30,  69,
    101,  51,   1,  33,  72,  17,  56,  88,   9,  41,
     80,  25,  64,  96,  49,   5,  37,  76,  21,  60,
     92,  13,  45,  84,  29,  68, 100,  52,   3,  35,
     74,  19,  58,  90,  11,  43,  82,  27,  66,  98,
     54,   7,  39,  78,  23,  62,  94,  15,  47,  86,
     31,  70, 102, 
    // Size 107
      0,  32,  75,  16,  59,  91,  48,   8,  40,  83,
     24,  67,  99,   4,  36,  79,  20,  63,  95,  55,
     12,  44,  87,  28,  71, 103,  52,   2,  34,  77,
     18,  61,  93,  50,  10,  42,  85,  26,  69, 101,
      6,  38,  81,  22,  65,  97,  57,  14,  46,  89,
     30,  73, 105,  53,   1,  33,  76,  17,  60,  92,
     49,   9,  41,  84,  25,  68, 100,   5,  37,  80,
     21,  64,  96,  56,  13,  45,  88,  29,  72, 104,
     54,   3,  35,  78,  19,  62,  94,  51,  11,  43,
     86,  27,  70, 102,   7,  39,  82,  23,  66,  98,
     58,  15,  47,  90,  31,  74, 106, 
    // Size 109
      0,  32,  77,  16,  61,  93,  48,   8,  40,  85,
     24,  69, 101,  52,   4,  36,  81,  20,  65,  97,
     57,  12,  44,  89,  28,  73, 105,   2,  34,  79,
     18,  63,  95,  50,  10,  42,  87,  26,  71, 103,
     55,   6,  38,  83,  22,  67,  99,  59,  14,  46,
     91,  30,  75, 107,  54,   1,  33,  78,  17,  62,
     94,  49,   9,  41,  86,  25,  70, 102,  53,   5,
     37,  82,  21,  66,  98,  58,  13,  45,  90,  29,
     74, 106,   3,  35,  80,  19,  64,  96,  51,  11,
     43,  88,  27,  72, 104,  56,   7,  39,  84,  23,
     68, 100,  60,  15,  47,  92,  31,  76, 108, 
    // Size 113
      0,  32,  81,  48,  16,  65,  97,   8,  40,  89,
     57,  24,  73, 105,   4,  36,  85,  52,  20,  69,
    101,  12,  44,  93,  61,  28,  77, 109,   2,  34,
     83,  50,  18,  67,  99,  10,  42,  91,  59,  26,
     75, 107,   6,  38,  87,  54,  22,  71, 103,  14,
     46,  95,  63,  30,  79, 111,  56,   1,  33,  82,
     49,  17,  66,  98,   9,  41,  90,  58,  25,  74,
    106,   5,  37,  86,  53,  21,  70, 102,  13,  45,
     94,  62,  29,  78, 110,   3,  35,  84,  51,  19,
     68, 100,  11,  43,  92,  60,  27,  76, 108,   7,
     39,  88,  55,  23,  72, 104,  15,  47,  96,  64,
     31,  80, 112, 
    // Size 127
      0,  32,  95,  48,  16,  79, 111,  56,   8,  40,
    103,  71,  24,  87, 119,  60,   4,  36,  99,  52,
     20,  83, 115,  67,  12,  44, 107,  75,  28,  91,
    123,  62,   2,  34,  97,  50,  18,  81, 113,  58,
     10,  42, 105,  73,  26,  89, 121,  65,   6,  38,
    101,  54,  22,  85, 117,  69,  14,  46, 109,  77,
     30,  93, 125,  63,   1,  33,  96,  49,  17,  80,
    112,  57,   9,  41, 104,  72,  25,  88, 120,  61,
      5,  37, 100,  53,  21,  84, 116,  68,  13,  45,
    108,  76,  29,  92, 124,  64,   3,  35,  98,  51,
     19,  82, 114,  59,  11,  43, 106,  74,  27,  90,
    122,  66,   7,  39, 102,  55,  23,  86, 118,  70,
     15,  47, 110,  78,  31,  94, 126, 
    // Size 131
      0,  67,  32,  99,  16,  83,  48, 115,   8,  75,
     40, 107,  24,  91,  56, 123,   4,  71,  36, 103,
     20,  87,  52, 119,  12,  79,  44, 111,  28,  95,
     60, 127,  64,   2,  69,  34, 101,  18,  85,  50,
    117,  10,  77,  42, 109,  26,  93,  58, 125,   6,
     73,  38, 105,  22,  89,  54, 121,  14,  81,  46,
    113,  30,  97,  62, 129,  65,   1,  68,  33, 100,
     17,  84,  49, 116,   9,  76,  41, 108,  25,  92,
     57, 124,   5,  72,  37, 104,  21,  88,  53, 120,
     13,  80,  45, 112,  29,  96,  61, 128,  66,   3,
     70,  35, 102,  19,  86,  51, 118,  11,  78,  43,
    110,  27,  94,  59, 126,   7,  74,  39, 106,  23,
     90,  55, 122,  15,  82,  47, 114,  31,  98,  63,
    130, 
    // Size 137
      0,  73,  32, 105,  16,  89,  48, 121,  64,   8,
     81,  40, 113,  24,  97,  56, 129,   4,  77,  36,
    109,  20,  93,  52, 125,  69,  12,  85,  44, 117,
     28, 101,  60, 133,   2,  75,  34, 107,  18,  91,
     50, 123,  66,  10,  83,  42, 115,  26,  99,  58,
    131,   6,  79,  38, 111,  22,  95,  54, 127,  71,
     14,  87,  46, 119,  30, 103,  62, 135,  68,   1,
     74,  33, 106,  17,  90,  49, 122,  65,   9,  82,
     41, 114,  25,  98,  57, 130,   5,  78,  37, 110,
     21,  94,  53, 126,  70,  13,  86,  45, 118,  29,
    102,  61, 134,   3,  76,  35, 108,  19,  92,  51,
    124,  67,  11,  84,  43, 116,  27, 100,  59, 132,
      7,  80,  39, 112,  23,  96,  55, 128,  72,  15,
     88,  47, 120,  31, 104,  63, 136, 
    // Size 139
      0,  75,  32, 107,  16,  91,  48, 123,  64,   8,
     83,  40, 115,  24,  99,  56, 131,   4,  79,  36,
    111,  20,  95,  52, 127,  71,  12,  87,  44, 119,
     28, 103,  60, 135,  68,   2,  77,  34, 109,  18,
     93,  50, 125,  66,  10,  85,  42, 117,  26, 101,
     58, 133,   6,  81,  38, 113,  22,  97,  54, 129,
     73,  14,  89,  46, 121,  30, 105,  62, 137,  69,
      1,  76,  33, 108,  17,  92,  49, 124,  65,   9,
     84,  41, 116,  25, 100,  57, 132,   5,  80,  37,
    112,  21,  96,  53, 128,  72,  13,  88,  45, 120,
     29, 104,  61, 136,  70,   3,  78,  35, 110,  19,
     94,  51, 126,  67,  11,  86,  43, 118,  27, 102,
     59, 134,   7,  82,  39, 114,  23,  98,  55, 130,
     74,  15,  90,  47, 122,  31, 106,  63, 138, 
    // Size 149
      0,  85,  32, 117,  64,  16, 101,  48, 133,   8,
     93,  40, 125,  77,  24, 109,  56, 141,  72,   4,
     89,  36, 121,  68,  20, 105,  52, 137,  12,  97,
     44, 129,  81,  28, 113,  60, 145,   2,  87,  34,
    119,  66,  18, 103,  50, 135,  10,  95,  42, 127,
     79,  26, 111,  58, 143,  75,   6,  91,  38, 123,
     70,  22, 107,  54, 139,  14,  99,  46, 131,  83,
     30, 115,  62, 147,  74,   1,  86,  33, 118,  65,
     17, 102,  49, 134,   9,  94,  41, 126,  78,  25,
    110,  57, 142,  73,   5,  90,  37, 122,  69,  21,
    106,  53, 138,  13,  98,  45, 130,  82,  29, 114,
     61, 146,   3,  88,  35, 120,  67,  19, 104,  51,
    136,  11,  96,  43, 128,  80,  27, 112,  59, 144,
     76,   7,  92,  39, 124,  71,  23, 108,  55, 140,
     15, 100,  47, 132,  84,  31, 116,  63, 148, 
    // Size 151
      0,  87,  32, 119,  64,  16, 103,  48, 135,   8,
     95,  40, 127,  79,  24, 111,  56, 143,  72,   4,
     91,  36, 123,  68,  20, 107,  52, 139,  12,  99,
     44, 131,  83,  28, 115,  60, 147,  74,   2,  89,
     34, 121,  66,  18, 105,  50, 137,  10,  97,  42,
    129,  81,  26, 113,  58, 145,  77,   6,  93,  38,
    125,  70,  22, 109,  54, 141,  14, 101,  46, 133,
     85,  30, 117,  62, 149,  75,   1,  88,  33, 120,
     65,  17, 104,  49, 136,   9,  96,  41, 128,  80,
     25, 112,  57, 144,  73,   5,  92,  37, 124,  69,
     21, 108,  53, 140,  13, 100,  45, 132,  84,  29,
    116,  61, 148,  76,   3,  90,  35, 122,  67,  19,
    106,  51, 138,  11,  98,  43, 130,  82,  27, 114,
     59, 146,  78,   7,  94,  39, 126,  71,  23, 110,
     55, 142,  15, 102,  47, 134,  86,  31, 118,  63,
    150, 
    // Size 157
      0,  93,  32, 125,  64,  16, 109,  48, 141,  72,
      8, 101,  40, 133,  85,  24, 117,  56, 149,  76,
      4,  97,  36, 129,  68,  20, 113,  52, 145,  81,
     12, 105,  44, 137,  89,  28, 121,  60, 153,   2,
     95,  34, 127,  66,  18, 111,  50, 143,  74,  10,
    103,  42, 135,  87,  26, 119,  58, 151,  79,   6,
     99,  38, 131,  70,  22, 115,  54, 147,  83,  14,
    107,  46, 139,  91,  30, 123,  62, 155,  78,   1,
     94,  33, 126,  65,  17, 110,  49, 142,  73,   9,
    102,  41, 134,  86,  25, 118,  57, 150,  77,   5,
     98,  37, 130,  69,  21, 114,  53, 146,  82,  13,
    106,  45, 138,  90,  29, 122,  61, 154,   3,  96,
     35, 128,  67,  19, 112,  51, 144,  75,  11, 104,
     43, 136,  88,  27, 120,  59, 152,  80,   7, 100,
     39, 132,  71,  23, 116,  55, 148,  84,  15, 108,
     47, 140,  92,  31, 124,  63, 156, 
    // Size 163
      0,  99,  64,  32, 131,  16, 115,  83,  48, 147,
      8, 107,  72,  40, 139,  24, 123,  91,  56, 155,
      4, 103,  68,  36, 135,  20, 119,  87,  52, 151,
     12, 111,  76,  44, 143,  28, 127,  95,  60, 159,
     80,   2, 101,  66,  34, 133,  18, 117,  85,  50,
    149,  10, 109,  74,  42, 141,  26, 125,  93,  58,
    157,   6, 105,  70,  38, 137,  22, 121,  89,  54,
    153,  14, 113,  78,  46, 145,  30, 129,  97,  62,
    161,  81,   1, 100,  65,  33, 132,  17, 116,  84,
     49, 148,   9, 108,  73,  41, 140,  25, 124,  92,
     57, 156,   5, 104,  69,  37, 136,  21, 120,  88,
     53, 152,  13, 112,  77,  45, 144,  29, 128,  96,
     61, 160,  82,   3, 102,  67,  35, 134,  19, 118,
     86,  51, 150,  11, 110,  75,  43, 142,  27, 126,
     94,  59, 158,   7, 106,  71,  39, 138,  23, 122,
     90,  55, 154,  15, 114,  79,  47, 146,  31, 130,
     98,  63, 162, 
    // Size 167
      0, 103,  64,  32, 135,  16, 119,  87,  48, 151,
      8, 111,  72,  40, 143,  24, 127,  95,  56, 159,
     80,   4, 107,  68,  36, 139,  20, 123,  91,  52,
    155,  12, 115,  76,  44, 147,  28, 131,  99,  60,
    163,  82,   2, 105,  66,  34, 137,  18, 121,  89,
     50, 153,  10, 113,  74,  42, 145,  26, 129,  97,
     58, 161,  85,   6, 109,  70,  38, 141,  22, 125,
     93,  54, 157,  14, 117,  78,  46, 149,  30, 133,
    101,  62, 165,  83,   1, 104,  65,  33, 136,  17,
    120,  88,  49, 152,   9, 112,  73,  41, 144,  25,
    128,  96,  57, 160,  81,   5, 108,  69,  37, 140,
     21, 124,  92,  53, 156,  13, 116,  77,  45, 148,
     29, 132, 100,  61, 164,  84,   3, 106,  67,  35,
    138,  19, 122,  90,  51, 154,  11, 114,  75,  43,
    146,  27, 130,  98,  59, 162,  86,   7, 110,  71,
     39, 142,  23, 126,  94,  55, 158,  15, 118,  79,
     47, 150,  31, 134, 102,  63, 166, 
    // Size 173
      0, 109,  64,  32, 141,  16, 125,  93,  48, 157,
     80,   8, 117,  72,  40, 149,  24, 133, 101,  56,
    165,  84,   4, 113,  68,  36, 145,  20, 129,  97,
     52, 161,  89,  12, 121,  76,  44, 153,  28, 137,
    105,  60, 169,   2, 111,  66,  34, 143,  18, 127,
     95,  50, 159,  82,  10, 119,  74,  42, 151,  26,
    135, 103,  58, 167,  87,   6, 115,  70,  38, 147,
     22, 131,  99,  54, 163,  91,  14, 123,  78,  46,
    155,  30, 139, 107,  62, 171,  86,   1, 110,  65,
     33, 142,  17, 126,  94,  49, 158,  81,   9, 118,
     73,  41, 150,  25, 134, 102,  57, 166,  85,   5,
    114,  69,  37, 146,  21, 130,  98,  53, 162,  90,
     13, 122,  77,  45, 154,  29, 138, 106,  61, 170,
      3, 112,  67,  35, 144,  19, 128,  96,  51, 160,
     83,  11, 120,  75,  43, 152,  27, 136, 104,  59,
    168,  88,   7, 116,  71,  39, 148,  23, 132, 100,
     55, 164,  92,  15, 124,  79,  47, 156,  31, 140,
    108,  63, 172, 
    // Size 179
      0, 115,  64,  32, 147,  80,  16, 131,  99,  48,
    163,   8, 123,  72,  40, 155,  91,  24, 139, 107,
     56, 171,   4, 119,  68,  36, 151,  84,  20, 135,
    103,  52, 167,  12, 127,  76,  44, 159,  95,  28,
    143, 111,  60, 175,  88,   2, 117,  66,  34, 149,
     82,  18, 133, 101,  50, 165,  10, 125,  74,  42,
    157,  93,  26, 141, 109,  58, 173,   6, 121,  70,
     38, 153,  86,  22, 137, 105,  54, 169,  14, 129,
     78,  46, 161,  97,  30, 145, 113,  62, 177,  89,
      1, 116,  65,  33, 148,  81,  17, 132, 100,  49,
    164,   9, 124,  73,  41, 156,  92,  25, 140, 108,
     57, 172,   5, 120,  69,  37, 152,  85,  21, 136,
    104,  53, 168,  13, 128,  77,  45, 160,  96,  29,
    144, 112,  61, 176,  90,   3, 118,  67,  35, 150,
     83,  19, 134, 102,  51, 166,  11, 126,  75,  43,
    158,  94,  27, 142, 110,  59, 174,   7, 122,  71,
     39, 154,  87,  23, 138, 106,  55, 170,  15, 130,
     79,  47, 162,  98,  31, 146, 114,  63, 178, 
    // Size 181
      0, 117,  64,  32, 149,  80,  16, 133, 101,  48,
    165,   8, 125,  72,  40, 157,  93,  24, 141, 109,
     56, 173,  88,   4, 121,  68,  36, 153,  84,  20,
    137, 105,  52, 169,  12, 129,  76,  44, 161,  97,
     28, 145, 113,  60, 177,   2, 119,  66,  34, 151,
     82,  18, 135, 103,  50, 167,  10, 127,  74,  42,
    159,  95,  26, 143, 111,  58, 175,  91,   6, 123,
     70,  38, 155,  86,  22, 139, 107,  54, 171,  14,
    131,  78,  46, 163,  99,  30, 147, 115,  62, 179,
     90,   1, 118,  65,  33, 150,  81,  17, 134, 102,
     49, 166,   9, 126,  73,  41, 158,  94,  25, 142,
    110,  57, 174,  89,   5, 122,  69,  37, 154,  85,
     21, 138, 106,  53, 170,  13, 130,  77,  45, 162,
     98,  29, 146, 114,  61, 178,   3, 120,  67,  35,
    152,  83,  19, 136, 104,  51, 168,  11, 128,  75,
     43, 160,  96,  27, 144, 112,  59, 176,  92,   7,
    124,  71,  39, 156,  87,  23, 140, 108,  55, 172,
     15, 132,  79,  47, 164, 100,  31, 148, 116,  63,
    180, 
    // Size 191
      0, 127,  64,  32, 159,  80,  16, 143, 111,  48,
    175,  88,   8, 135,  72,  40, 167, 103,  24, 151,
    119,  56, 183,  92,   4, 131,  68,  36, 163,  84,
     20, 147, 115,  52, 179,  99,  12, 139,  76,  44,
    171, 107,  28, 155, 123,  60, 187,  94,   2, 129,
     66,  34, 161,  82,  18, 145, 113,  50, 177,  90,
     10, 137,  74,  42, 169, 105,  26, 153, 121,  58,
    185,  97,   6, 133,  70,  38, 165,  86,  22, 149,
    117,  54, 181, 101,  14, 141,  78,  46, 173, 109,
     30, 157, 125,  62, 189,  95,   1, 128,  65,  33,
    160,  81,  17, 144, 112,  49, 176,  89,   9, 136,
     73,  41, 168, 104,  25, 152, 120,  57, 184,  93,
      5, 132,  69,  37, 164,  85,  21, 148, 116,  53,
    180, 100,  13, 140,  77,  45, 172, 108,  29, 156,
    124,  61, 188,  96,   3, 130,  67,  35, 162,  83,
     19, 146, 114,  51, 178,  91,  11, 138,  75,  43,
    170, 106,  27, 154, 122,  59, 186,  98,   7, 134,
     71,  39, 166,  87,  23, 150, 118,  55, 182, 102,
     15, 142,  79,  47, 174, 110,  31, 158, 126,  63,
    190, 
    // Size 193
      0,  64, 129,  32,  97, 161,  16,  80, 145,  48,
    113, 177,   8,  72, 137,  40, 105, 169,  24,  88,
    153,  56, 121, 185,   4,  68, 133,  36, 101, 165,
     20,  84, 149,  52, 117, 181,  12,  76, 141,  44,
    109, 173,  28,  92, 157,  60, 125, 189,   2,  66,
    131,  34,  99, 163,  18,  82, 147,  50, 115, 179,
     10,  74, 139,  42, 107, 171,  26,  90, 155,  58,
    123, 187,   6,  70, 135,  38, 103, 167,  22,  86,
    151,  54, 119, 183,  14,  78, 143,  46, 111, 175,
     30,  94, 159,  62, 127, 191,  96,   1,  65, 130,
     33,  98, 162,  17,  81, 146,  49, 114, 178,   9,
     73, 138,  41, 106, 170,  25,  89, 154,  57, 122,
    186,   5,  69, 134,  37, 102, 166,  21,  85, 150,
     53, 118, 182,  13,  77, 142,  45, 110, 174,  29,
     93, 158,  61, 126, 190,   3,  67, 132,  35, 100,
    164,  19,  83, 148,  51, 116, 180,  11,  75, 140,
     43, 108, 172,  27,  91, 156,  59, 124, 188,   7,
     71, 136,  39, 104, 168,  23,  87, 152,  55, 120,
    184,  15,  79, 144,  47, 112, 176,  31,  95, 160,
     63, 128, 192, 
    // Size 197
      0,  64, 133,  32, 101, 165,  16,  80, 149,  48,
    117, 181,   8,  72, 141,  40, 109, 173,  24,  88,
    157,  56, 125, 189,  96,   4,  68, 137,  36, 105,
    169,  20,  84, 153,  52, 121, 185,  12,  76, 145,
     44, 113, 177,  28,  92, 161,  60, 129, 193,   2,
     66, 135,  34, 103, 167,  18,  82, 151,  50, 119,
    183,  10,  74, 143,  42, 111, 175,  26,  90, 159,
     58, 127, 191,  99,   6,  70, 139,  38, 107, 171,
     22,  86, 155,  54, 123, 187,  14,  78, 147,  46,
    115, 179,  30,  94, 163,  62, 131, 195,  98,   1,
     65, 134,  33, 102, 166,  17,  81, 150,  49, 118,
    182,   9,  73, 142,  41, 110, 174,  25,  89, 158,
     57, 126, 190,  97,   5,  69, 138,  37, 106, 170,
     21,  85, 154,  53, 122, 186,  13,  77, 146,  45,
    114, 178,  29,  93, 162,  61, 130, 194,   3,  67,
    136,  35, 104, 168,  19,  83, 152,  51, 120, 184,
     11,  75, 144,  43, 112, 176,  27,  91, 160,  59,
    128, 192, 100,   7,  71, 140,  39, 108, 172,  23,
     87, 156,  55, 124, 188,  15,  79, 148,  47, 116,
    180,  31,  95, 164,  63, 132, 196, 
    // Size 199
      0,  64, 135,  32, 103, 167,  16,  80, 151,  48,
    119, 183,   8,  72, 143,  40, 111, 175,  24,  88,
    159,  56, 127, 191,  96,   4,  68, 139,  36, 107,
    171,  20,  84, 155,  52, 123, 187,  12,  76, 147,
     44, 115, 179,  28,  92, 163,  60, 131, 195,  98,
      2,  66, 137,  34, 105, 169,  18,  82, 153,  50,
    121, 185,  10,  74, 145,  42, 113, 177,  26,  90,
    161,  58, 129, 193, 101,   6,  70, 141,  38, 109,
    173,  22,  86, 157,  54, 125, 189,  14,  78, 149,
     46, 117, 181,  30,  94, 165,  62, 133, 197,  99,
      1,  65, 136,  33, 104, 168,  17,  81, 152,  49,
    120, 184,   9,  73, 144,  41, 112, 176,  25,  89,
    160,  57, 128, 192,  97,   5,  69, 140,  37, 108,
    172,  21,  85, 156,  53, 124, 188,  13,  77, 148,
     45, 116, 180,  29,  93, 164,  61, 132, 196, 100,
      3,  67, 138,  35, 106, 170,  19,  83, 154,  51,
    122, 186,  11,  75, 146,  43, 114, 178,  27,  91,
    162,  59, 130, 194, 102,   7,  71, 142,  39, 110,
    174,  23,  87, 158,  55, 126, 190,  15,  79, 150,
     47, 118, 182,  31,  95, 166,  63, 134, 198, 
    // Size 211
      0,  64, 147,  32, 115, 179,  96,  16,  80, 163,
     48, 131, 195,   8,  72, 155,  40, 123, 187, 107,
     24,  88, 171,  56, 139, 203,   4,  68, 151,  36,
    119, 183, 100,  20,  84, 167,  52, 135, 199,  12,
     76, 159,  44, 127, 191, 111,  28,  92, 175,  60,
    143, 207, 104,   2,  66, 149,  34, 117, 181,  98,
     18,  82, 165,  50, 133, 197,  10,  74, 157,  42,
    125, 189, 109,  26,  90, 173,  58, 141, 205,   6,
     70, 153,  38, 121, 185, 102,  22,  86, 169,  54,
    137, 201,  14,  78, 161,  46, 129, 193, 113,  30,
     94, 177,  62, 145, 209, 105,   1,  65, 148,  33,
    116, 180,  97,  17,  81, 164,  49, 132, 196,   9,
     73, 156,  41, 124, 188, 108,  25,  89, 172,  57,
    140, 204,   5,  69, 152,  37, 120, 184, 101,  21,
     85, 168,  53, 136, 200,  13,  77, 160,  45, 128,
    192, 112,  29,  93, 176,  61, 144, 208, 106,   3,
     67, 150,  35, 118, 182,  99,  19,  83, 166,  51,
    134, 198,  11,  75, 158,  43, 126, 190, 110,  27,
     91, 174,  59, 142, 206,   7,  71, 154,  39, 122,
    186, 103,  23,  87, 170,  55, 138, 202,  15,  79,
    162,  47, 130, 194, 114,  31,  95, 178,  63, 146,
    210, 
    // Size 223
      0,  64, 159,  32, 127, 191,  96,  16,  80, 175,
     48, 143, 207, 104,   8,  72, 167,  40, 135, 199,
    119,  24,  88, 183,  56, 151, 215, 108,   4,  68,
    163,  36, 131, 195, 100,  20,  84, 179,  52, 147,
    211, 115,  12,  76, 171,  44, 139, 203, 123,  28,
     92, 187,  60, 155, 219, 110,   2,  66, 161,  34,
    129, 193,  98,  18,  82, 177,  50, 145, 209, 106,
     10,  74, 169,  42, 137, 201, 121,  26,  90, 185,
     58, 153, 217, 113,   6,  70, 165,  38, 133, 197,
    102,  22,  86, 181,  54, 149, 213, 117,  14,  78,
    173,  46, 141, 205, 125,  30,  94, 189,  62, 157,
    221, 111,   1,  65, 160,  33, 128, 192,  97,  17,
     81, 176,  49, 144, 208, 105,   9,  73, 168,  41,
    136, 200, 120,  25,  89, 184,  57, 152, 216, 109,
      5,  69, 164,  37, 132, 196, 101,  21,  85, 180,
     53, 148, 212, 116,  13,  77, 172,  45, 140, 204,
    124,  29,  93, 188,  61, 156, 220, 112,   3,  67,
    162,  35, 130, 194,  99,  19,  83, 178,  51, 146,
    210, 107,  11,  75, 170,  43, 138, 202, 122,  27,
     91, 186,  59, 154, 218, 114,   7,  71, 166,  39,
    134, 198, 103,  23,  87, 182,  55, 150, 214, 118,
     15,  79, 174,  47, 142, 206, 126,  31,  95, 190,
     63, 158, 222, 
    // Size 227
      0,  64, 163,  96,  32, 131, 195,  16,  80, 179,
    115,  48, 147, 211,   8,  72, 171, 104,  40, 139,
    203,  24,  88, 187, 123,  56, 155, 219,   4,  68,
    167, 100,  36, 135, 199,  20,  84, 183, 119,  52,
    151, 215,  12,  76, 175, 108,  44, 143, 207,  28,
     92, 191, 127,  60, 159, 223, 112,   2,  66, 165,
     98,  34, 133, 197,  18,  82, 181, 117,  50, 149,
    213,  10,  74, 173, 106,  42, 141, 205,  26,  90,
    189, 125,  58, 157, 221,   6,  70, 169, 102,  38,
    137, 201,  22,  86, 185, 121,  54, 153, 217,  14,
     78, 177, 110,  46, 145, 209,  30,  94, 193, 129,
     62, 161, 225, 113,   1,  65, 164,  97,  33, 132,
    196,  17,  81, 180, 116,  49, 148, 212,   9,  73,
    172, 105,  41, 140, 204,  25,  89, 188, 124,  57,
    156, 220,   5,  69, 168, 101,  37, 136, 200,  21,
     85, 184, 120,  53, 152, 216,  13,  77, 176, 109,
     45, 144, 208,  29,  93, 192, 128,  61, 160, 224,
    114,   3,  67, 166,  99,  35, 134, 198,  19,  83,
    182, 118,  51, 150, 214,  11,  75, 174, 107,  43,
    142, 206,  27,  91, 190, 126,  59, 158, 222,   7,
     71, 170, 103,  39, 138, 202,  23,  87, 186, 122,
     55, 154, 218,  15,  79, 178, 111,  47, 146, 210,
     31,  95, 194, 130,  63, 162, 226, 
    // Size 229
      0,  64, 165,  96,  32, 133, 197,  16,  80, 181,
    117,  48, 149, 213,   8,  72, 173, 104,  40, 141,
    205,  24,  88, 189, 125,  56, 157, 221, 112,   4,
     68, 169, 100,  36, 137, 201,  20,  84, 185, 121,
     52, 153, 217,  12,  76, 177, 108,  44, 145, 209,
     28,  92, 193, 129,  60, 161, 225,   2,  66, 167,
     98,  34, 135, 199,  18,  82, 183, 119,  50, 151,
    215,  10,  74, 175, 106,  42, 143, 207,  26,  90,
    191, 127,  58, 159, 223, 115,   6,  70, 171, 102,
     38, 139, 203,  22,  86, 187, 123,  54, 155, 219,
     14,  78, 179, 110,  46, 147, 211,  30,  94, 195,
    131,  62, 163, 227, 114,   1,  65, 166,  97,  33,
    134, 198,  17,  81, 182, 118,  49, 150, 214,   9,
     73, 174, 105,  41, 142, 206,  25,  89, 190, 126,
     57, 158, 222, 113,   5,  69, 170, 101,  37, 138,
    202,  21,  85, 186, 122,  53, 154, 218,  13,  77,
    178, 109,  45, 146, 210,  29,  93, 194, 130,  61,
    162, 226,   3,  67, 168,  99,  35, 136, 200,  19,
     83, 184, 120,  51, 152, 216,  11,  75, 176, 107,
     43, 144, 208,  27,  91, 192, 128,  59, 160, 224,
    116,   7,  71, 172, 103,  39, 140, 204,  23,  87,
    188, 124,  55, 156, 220,  15,  79, 180, 111,  47,
    148, 212,  31,  95, 196, 132,  63, 164, 228, 
    // Size 233
      0,  64, 169,  96,  32, 137, 201,  16,  80, 185,
    121,  48, 153, 217, 112,   8,  72, 177, 104,  40,
    145, 209,  24,  88, 193, 129,  56, 161, 225,   4,
     68, 173, 100,  36, 141, 205,  20,  84, 189, 125,
     52, 157, 221, 117,  12,  76, 181, 108,  44, 149,
    213,  28,  92, 197, 133,  60, 165, 229,   2,  66,
    171,  98,  34, 139, 203,  18,  82, 187, 123,  50,
    155, 219, 114,  10,  74, 179, 106,  42, 147, 211,
     26,  90, 195, 131,  58, 163, 227,   6,  70, 175,
    102,  38, 143, 207,  22,  86, 191, 127,  54, 159,
    223, 119,  14,  78, 183, 110,  46, 151, 215,  30,
     94, 199, 135,  62, 167, 231, 116,   1,  65, 170,
     97,  33, 138, 202,  17,  81, 186, 122,  49, 154,
    218, 113,   9,  73, 178, 105,  41, 146, 210,  25,
     89, 194, 130,  57, 162, 226,   5,  69, 174, 101,
     37, 142, 206,  21,  85, 190, 126,  53, 158, 222,
    118,  13,  77, 182, 109,  45, 150, 214,  29,  93,
    198, 134,  61, 166, 230,   3,  67, 172,  99,  35,
    140, 204,  19,  83, 188, 124,  51, 156, 220, 115,
     11,  75, 180, 107,  43, 148, 212,  27,  91, 196,
    132,  59, 164, 228,   7,  71, 176, 103,  39, 144,
    208,  23,  87, 192, 128,  55, 160, 224, 120,  15,
     79, 184, 111,  47, 152, 216,  31,  95, 200, 136,
     63, 168, 232, 
    // Size 239
      0,  64, 175,  96,  32, 143, 207,  16,  80, 191,
    127,  48, 159, 223, 112,   8,  72, 183, 104,  40,
    151, 215,  24,  88, 199, 135,  56, 167, 231, 116,
      4,  68, 179, 100,  36, 147, 211,  20,  84, 195,
    131,  52, 163, 227, 123,  12,  76, 187, 108,  44,
    155, 219,  28,  92, 203, 139,  60, 171, 235, 118,
      2,  66, 177,  98,  34, 145, 209,  18,  82, 193,
    129,  50, 161, 225, 114,  10,  74, 185, 106,  42,
    153, 217,  26,  90, 201, 137,  58, 169, 233, 121,
      6,  70, 181, 102,  38, 149, 213,  22,  86, 197,
    133,  54, 165, 229, 125,  14,  78, 189, 110,  46,
    157, 221,  30,  94, 205, 141,  62, 173, 237, 119,
      1,  65, 176,  97,  33, 144, 208,  17,  81, 192,
    128,  49, 160, 224, 113,   9,  73, 184, 105,  41,
    152, 216,  25,  89, 200, 136,  57, 168, 232, 117,
      5,  69, 180, 101,  37, 148, 212,  21,  85, 196,
    132,  53, 164, 228, 124,  13,  77, 188, 109,  45,
    156, 220,  29,  93, 204, 140,  61, 172, 236, 120,
      3,  67, 178,  99,  35, 146, 210,  19,  83, 194,
    130,  51, 162, 226, 115,  11,  75, 186, 107,  43,
    154, 218,  27,  91, 202, 138,  59, 170, 234, 122,
      7,  71, 182, 103,  39, 150, 214,  23,  87, 198,
    134,  55, 166, 230, 126,  15,  79, 190, 111,  47,
    158, 222,  31,  95, 206, 142,  63, 174, 238, 
    // Size 241
      0,  64, 177,  96,  32, 145, 209, 112,  16,  80,
    193, 129,  48, 161, 225,   8,  72, 185, 104,  40,
    153, 217, 121,  24,  88, 201, 137,  56, 169, 233,
      4,  68, 181, 100,  36, 149, 213, 116,  20,  84,
    197, 133,  52, 165, 229,  12,  76, 189, 108,  44,
    157, 221, 125,  28,  92, 205, 141,  60, 173, 237,
      2,  66, 179,  98,  34, 147, 211, 114,  18,  82,
    195, 131,  50, 163, 227,  10,  74, 187, 106,  42,
    155, 219, 123,  26,  90, 203, 139,  58, 171, 235,
      6,  70, 183, 102,  38, 151, 215, 118,  22,  86,
    199, 135,  54, 167, 231,  14,  78, 191, 110,  46,
    159, 223, 127,  30,  94, 207, 143,  62, 175, 239,
    120,   1,  65, 178,  97,  33, 146, 210, 113,  17,
     81, 194, 130,  49, 162, 226,   9,  73, 186, 105,
     41, 154, 218, 122,  25,  89, 202, 138,  57, 170,
    234,   5,  69, 182, 101,  37, 150, 214, 117,  21,
     85, 198, 134,  53, 166, 230,  13,  77, 190, 109,
     45, 158, 222, 126,  29,  93, 206, 142,  61, 174,
    238,   3,  67, 180,  99,  35, 148, 212, 115,  19,
     83, 196, 132,  51, 164, 228,  11,  75, 188, 107,
     43, 156, 220, 124,  27,  91, 204, 140,  59, 172,
    236,   7,  71, 184, 103,  39, 152, 216, 119,  23,
     87, 200, 136,  55, 168, 232,  15,  79, 192, 111,
     47, 160, 224, 128,  31,  95, 208, 144,  63, 176,
    240, 
    // Size 251
      0,  64, 187,  96,  32, 155, 219, 112,  16,  80,
    203, 139,  48, 171, 235, 120,   8,  72, 195, 104,
     40, 163, 227, 131,  24,  88, 211, 147,  56, 179,
    243,   4,  68, 191, 100,  36, 159, 223, 116,  20,
     84, 207, 143,  52, 175, 239, 127,  12,  76, 199,
    108,  44, 167, 231, 135,  28,  92, 215, 151,  60,
    183, 247, 124,   2,  66, 189,  98,  34, 157, 221,
    114,  18,  82, 205, 141,  50, 173, 237, 122,  10,
     74, 197, 106,  42, 165, 229, 133,  26,  90, 213,
    149,  58, 181, 245,   6,  70, 193, 102,  38, 161,
    225, 118,  22,  86, 209, 145,  54, 177, 241, 129,
     14,  78, 201, 110,  46, 169, 233, 137,  30,  94,
    217, 153,  62, 185, 249, 125,   1,  65, 188,  97,
     33, 156, 220, 113,  17,  81, 204, 140,  49, 172,
    236, 121,   9,  73, 196, 105,  41, 164, 228, 132,
     25,  89, 212, 148,  57, 180, 244,   5,  69, 192,
    101,  37, 160, 224, 117,  21,  85, 208, 144,  53,
    176, 240, 128,  13,  77, 200, 109,  45, 168, 232,
    136,  29,  93, 216, 152,  61, 184, 248, 126,   3,
     67, 190,  99,  35, 158, 222, 115,  19,  83, 206,
    142,  51, 174, 238, 123,  11,  75, 198, 107,  43,
    166, 230, 134,  27,  91, 214, 150,  59, 182, 246,
      7,  71, 194, 103,  39, 162, 226, 119,  23,  87,
    210, 146,  55, 178, 242, 130,  15,  79, 202, 111,
     47, 170, 234, 138,  31,  95, 218, 154,  63, 186,
    250, 
    // Size 257
      0, 129,  64, 193,  32, 161,  96, 225,  16, 145,
     80, 209,  48, 177, 112, 241,   8, 137,  72, 201,
     40, 169, 104, 233,  24, 153,  88, 217,  56, 185,
    120, 249,   4, 133,  68, 197,  36, 165, 100, 229,
     20, 149,  84, 213,  52, 181, 116, 245,  12, 141,
     76, 205,  44, 173, 108, 237,  28, 157,  92, 221,
     60, 189, 124, 253,   2, 131,  66, 195,  34, 163,
     98, 227,  18, 147,  82, 211,  50, 179, 114, 243,
     10, 139,  74, 203,  42, 171, 106, 235,  26, 155,
     90, 219,  58, 187, 122, 251,   6, 135,  70, 199,
     38, 167, 102, 231,  22, 151,  86, 215,  54, 183,
    118, 247,  14, 143,  78, 207,  46, 175, 110, 239,
     30, 159,  94, 223,  62, 191, 126, 255, 128,   1,
    130,  65, 194,  33, 162,  97, 226,  17, 146,  81,
    210,  49, 178, 113, 242,   9, 138,  73, 202,  41,
    170, 105, 234,  25, 154,  89, 218,  57, 186, 121,
    250,   5, 134,  69, 198,  37, 166, 101, 230,  21,
    150,  85, 214,  53, 182, 117, 246,  13, 142,  77,
    206,  45, 174, 109, 238,  29, 158,  93, 222,  61,
    190, 125, 254,   3, 132,  67, 196,  35, 164,  99,
    228,  19, 148,  83, 212,  51, 180, 115, 244,  11,
    140,  75, 204,  43, 172, 107, 236,  27, 156,  91,
    220,  59, 188, 123, 252,   7, 136,  71, 200,  39,
    168, 103, 232,  23, 152,  87, 216,  55, 184, 119,
    248,  15, 144,  79, 208,  47, 176, 111, 240,  31,
    160,  95, 224,  63, 192, 127, 256, 
    // Size 263
      0, 135,  64, 199,  32, 167,  96, 231,  16, 151,
     80, 215,  48, 183, 112, 247,   8, 143,  72, 207,
     40, 175, 104, 239,  24, 159,  88, 223,  56, 191,
    120, 255, 128,   4, 139,  68, 203,  36, 171, 100,
    235,  20, 155,  84, 219,  52, 187, 116, 251,  12,
    147,  76, 211,  44, 179, 108, 243,  28, 163,  92,
    227,  60, 195, 124, 259, 130,   2, 137,  66, 201,
     34, 169,  98, 233,  18, 153,  82, 217,  50, 185,
    114, 249,  10, 145,  74, 209,  42, 177, 106, 241,
     26, 161,  90, 225,  58, 193, 122, 257, 133,   6,
    141,  70, 205,  38, 173, 102, 237,  22, 157,  86,
    221,  54, 189, 118, 253,  14, 149,  78, 213,  46,
    181, 110, 245,  30, 165,  94, 229,  62, 197, 126,
    261, 131,   1, 136,  65, 200,  33, 168,  97, 232,
     17, 152,  81, 216,  49, 184, 113, 248,   9, 144,
     73, 208,  41, 176, 105, 240,  25, 160,  89, 224,
     57, 192, 121, 256, 129,   5, 140,  69, 204,  37,
    172, 101, 236,  21, 156,  85, 220,  53, 188, 117,
    252,  13, 148,  77, 212,  45, 180, 109, 244,  29,
    164,  93, 228,  61, 196, 125, 260, 132,   3, 138,
     67, 202,  35, 170,  99, 234,  19, 154,  83, 218,
     51, 186, 115, 250,  11, 146,  75, 210,  43, 178,
    107, 242,  27, 162,  91, 226,  59, 194, 123, 258,
    134,   7, 142,  71, 206,  39, 174, 103, 238,  23,
    158,  87, 222,  55, 190, 119, 254,  15, 150,  79,
    214,  47, 182, 111, 246,  31, 166,  95, 230,  63,
    198, 127, 262, 
    // Size 269
      0, 141,  64, 205,  32, 173,  96, 237,  16, 157,
     80, 221,  48, 189, 112, 253, 128,   8, 149,  72,
    213,  40, 181, 104, 245,  24, 165,  88, 229,  56,
    197, 120, 261, 132,   4, 145,  68, 209,  36, 177,
    100, 241,  20, 161,  84, 225,  52, 193, 116, 257,
    137,  12, 153,  76, 217,  44, 185, 108, 249,  28,
    169,  92, 233,  60, 201, 124, 265,   2, 143,  66,
    207,  34, 175,  98, 239,  18, 159,  82, 223,  50,
    191, 114, 255, 130,  10, 151,  74, 215,  42, 183,
    106, 247,  26, 167,  90, 231,  58, 199, 122, 263,
    135,   6, 147,  70, 211,  38, 179, 102, 243,  22,
    163,  86, 227,  54, 195, 118, 259, 139,  14, 155,
     78, 219,  46, 187, 110, 251,  30, 171,  94, 235,
     62, 203, 126, 267, 134,   1, 142,  65, 206,  33,
    174,  97, 238,  17, 158,  81, 222,  49, 190, 113,
    254, 129,   9, 150,  73, 214,  41, 182, 105, 246,
     25, 166,  89, 230,  57, 198, 121, 262, 133,   5,
    146,  69, 210,  37, 178, 101, 242,  21, 162,  85,
    226,  53, 194, 117, 258, 138,  13, 154,  77, 218,
     45, 186, 109, 250,  29, 170,  93, 234,  61, 202,
    125, 266,   3, 144,  67, 208,  35, 176,  99, 240,
     19, 160,  83, 224,  51, 192, 115, 256, 131,  11,
    152,  75, 216,  43, 184, 107, 248,  27, 168,  91,
    232,  59, 200, 123, 264, 136,   7, 148,  71, 212,
     39, 180, 103, 244,  23, 164,  87, 228,  55, 196,
    119, 260, 140,  15, 156,  79, 220,  47, 188, 111,
    252,  31, 172,  95, 236,  63, 204, 127, 268, 
    // Size 271
      0, 143,  64, 207,  32, 175,  96, 239,  16, 159,
     80, 223,  48, 191, 112, 255, 128,   8, 151,  72,
    215,  40, 183, 104, 247,  24, 167,  88, 231,  56,
    199, 120, 263, 132,   4, 147,  68, 211,  36, 179,
    100, 243,  20, 163,  84, 227,  52, 195, 116, 259,
    139,  12, 155,  76, 219,  44, 187, 108, 251,  28,
    171,  92, 235,  60, 203, 124, 267, 134,   2, 145,
     66, 209,  34, 177,  98, 241,  18, 161,  82, 225,
     50, 193, 114, 257, 130,  10, 153,  74, 217,  42,
    185, 106, 249,  26, 169,  90, 233,  58, 201, 122,
    265, 137,   6, 149,  70, 213,  38, 181, 102, 245,
     22, 165,  86, 229,  54, 197, 118, 261, 141,  14,
    157,  78, 221,  46, 189, 110, 253,  30, 173,  94,
    237,  62, 205, 126, 269, 135,   1, 144,  65, 208,
     33, 176,  97, 240,  17, 160,  81, 224,  49, 192,
    113, 256, 129,   9, 152,  73, 216,  41, 184, 105,
    248,  25, 168,  89, 232,  57, 200, 121, 264, 133,
      5, 148,  69, 212,  37, 180, 101, 244,  21, 164,
     85, 228,  53, 196, 117, 260, 140,  13, 156,  77,
    220,  45, 188, 109, 252,  29, 172,  93, 236,  61,
    204, 125, 268, 136,   3, 146,  67, 210,  35, 178,
     99, 242,  19, 162,  83, 226,  51, 194, 115, 258,
    131,  11, 154,  75, 218,  43, 186, 107, 250,  27,
    170,  91, 234,  59, 202, 123, 266, 138,   7, 150,
     71, 214,  39, 182, 103, 246,  23, 166,  87, 230,
     55, 198, 119, 262, 142,  15, 158,  79, 222,  47,
    190, 111, 254,  31, 174,  95, 238,  63, 206, 127,
    270, 
    // Size 277
      0, 149,  64, 213,  32, 181,  96, 245, 128,  16,
    165,  80, 229,  48, 197, 112, 261,   8, 157,  72,
    221,  40, 189, 104, 253, 141,  24, 173,  88, 237,
     56, 205, 120, 269, 136,   4, 153,  68, 217,  36,
    185, 100, 249, 132,  20, 169,  84, 233,  52, 201,
    116, 265,  12, 161,  76, 225,  44, 193, 108, 257,
    145,  28, 177,  92, 241,  60, 209, 124, 273,   2,
    151,  66, 215,  34, 183,  98, 247, 130,  18, 167,
     82, 231,  50, 199, 114, 263,  10, 159,  74, 223,
     42, 191, 106, 255, 143,  26, 175,  90, 239,  58,
    207, 122, 271, 139,   6, 155,  70, 219,  38, 187,
    102, 251, 134,  22, 171,  86, 235,  54, 203, 118,
    267,  14, 163,  78, 227,  46, 195, 110, 259, 147,
     30, 179,  94, 243,  62, 211, 126, 275, 138,   1,
    150,  65, 214,  33, 182,  97, 246, 129,  17, 166,
     81, 230,  49, 198, 113, 262,   9, 158,  73, 222,
     41, 190, 105, 254, 142,  25, 174,  89, 238,  57,
    206, 121, 270, 137,   5, 154,  69, 218,  37, 186,
    101, 250, 133,  21, 170,  85, 234,  53, 202, 117,
    266,  13, 162,  77, 226,  45, 194, 109, 258, 146,
     29, 178,  93, 242,  61, 210, 125, 274,   3, 152,
     67, 216,  35, 184,  99, 248, 131,  19, 168,  83,
    232,  51, 200, 115, 264,  11, 160,  75, 224,  43,
    192, 107, 256, 144,  27, 176,  91, 240,  59, 208,
    123, 272, 140,   7, 156,  71, 220,  39, 188, 103,
    252, 135,  23, 172,  87, 236,  55, 204, 119, 268,
     15, 164,  79, 228,  47, 196, 111, 260, 148,  31,
    180,  95, 244,  63, 212, 127, 276, 
    // Size 281
      0, 153,  64, 217,  32, 185,  96, 249, 128,  16,
    169,  80, 233,  48, 201, 112, 265, 136,   8, 161,
     72, 225,  40, 193, 104, 257, 145,  24, 177,  88,
    241,  56, 209, 120, 273,   4, 157,  68, 221,  36,
    189, 100, 253, 132,  20, 173,  84, 237,  52, 205,
    116, 269, 141,  12, 165,  76, 229,  44, 197, 108,
    261, 149,  28, 181,  92, 245,  60, 213, 124, 277,
      2, 155,  66, 219,  34, 187,  98, 251, 130,  18,
    171,  82, 235,  50, 203, 114, 267, 138,  10, 163,
     74, 227,  42, 195, 106, 259, 147,  26, 179,  90,
    243,  58, 211, 122, 275,   6, 159,  70, 223,  38,
    191, 102, 255, 134,  22, 175,  86, 239,  54, 207,
    118, 271, 143,  14, 167,  78, 231,  46, 199, 110,
    263, 151,  30, 183,  94, 247,  62, 215, 126, 279,
    140,   1, 154,  65, 218,  33, 186,  97, 250, 129,
     17, 170,  81, 234,  49, 202, 113, 266, 137,   9,
    162,  73, 226,  41, 194, 105, 258, 146,  25, 178,
     89, 242,  57, 210, 121, 274,   5, 158,  69, 222,
     37, 190, 101, 254, 133,  21, 174,  85, 238,  53,
    206, 117, 270, 142,  13, 166,  77, 230,  45, 198,
    109, 262, 150,  29, 182,  93, 246,  61, 214, 125,
    278,   3, 156,  67, 220,  35, 188,  99, 252, 131,
     19, 172,  83, 236,  51, 204, 115, 268, 139,  11,
    164,  75, 228,  43, 196, 107, 260, 148,  27, 180,
     91, 244,  59, 212, 123, 276,   7, 160,  71, 224,
     39, 192, 103, 256, 135,  23, 176,  87, 240,  55,
    208, 119, 272, 144,  15, 168,  79, 232,  47, 200,
    111, 264, 152,  31, 184,  95, 248,  63, 216, 127,
    280, 
    // Size 283
      0, 155,  64, 219,  32, 187,  96, 251, 128,  16,
    171,  80, 235,  48, 203, 112, 267, 136,   8, 163,
     72, 227,  40, 195, 104, 259, 147,  24, 179,  88,
    243,  56, 211, 120, 275,   4, 159,  68, 223,  36,
    191, 100, 255, 132,  20, 175,  84, 239,  52, 207,
    116, 271, 143,  12, 167,  76, 231,  44, 199, 108,
    263, 151,  28, 183,  92, 247,  60, 215, 124, 279,
    140,   2, 157,  66, 221,  34, 189,  98, 253, 130,
     18, 173,  82, 237,  50, 205, 114, 269, 138,  10,
    165,  74, 229,  42, 197, 106, 261, 149,  26, 181,
     90, 245,  58, 213, 122, 277,   6, 161,  70, 225,
     38, 193, 102, 257, 134,  22, 177,  86, 241,  54,
    209, 118, 273, 145,  14, 169,  78, 233,  46, 201,
    110, 265, 153,  30, 185,  94, 249,  62, 217, 126,
    281, 141,   1, 156,  65, 220,  33, 188,  97, 252,
    129,  17, 172,  81, 236,  49, 204, 113, 268, 137,
      9, 164,  73, 228,  41, 196, 105, 260, 148,  25,
    180,  89, 244,  57, 212, 121, 276,   5, 160,  69,
    224,  37, 192, 101, 256, 133,  21, 176,  85, 240,
     53, 208, 117, 272, 144,  13, 168,  77, 232,  45,
    200, 109, 264, 152,  29, 184,  93, 248,  61, 216,
    125, 280, 142,   3, 158,  67, 222,  35, 190,  99,
    254, 131,  19, 174,  83, 238,  51, 206, 115, 270,
    139,  11, 166,  75, 230,  43, 198, 107, 262, 150,
     27, 182,  91, 246,  59, 214, 123, 278,   7, 162,
     71, 226,  39, 194, 103, 258, 135,  23, 178,  87,
    242,  55, 210, 119, 274, 146,  15, 170,  79, 234,
     47, 202, 111, 266, 154,  31, 186,  95, 250,  63,
    218, 127, 282, 
    // Size 293
      0, 165,  64, 229, 128,  32, 197,  96, 261,  16,
    181,  80, 245, 149,  48, 213, 112, 277,   8, 173,
     72, 237, 136,  40, 205, 104, 269,  24, 189,  88,
    253, 157,  56, 221, 120, 285, 144,   4, 169,  68,
    233, 132,  36, 201, 100, 265,  20, 185,  84, 249,
    153,  52, 217, 116, 281,  12, 177,  76, 241, 140,
     44, 209, 108, 273,  28, 193,  92, 257, 161,  60,
    225, 124, 289,   2, 167,  66, 231, 130,  34, 199,
     98, 263,  18, 183,  82, 247, 151,  50, 215, 114,
    279,  10, 175,  74, 239, 138,  42, 207, 106, 271,
     26, 191,  90, 255, 159,  58, 223, 122, 287, 147,
      6, 171,  70, 235, 134,  38, 203, 102, 267,  22,
    187,  86, 251, 155,  54, 219, 118, 283,  14, 179,
     78, 243, 142,  46, 211, 110, 275,  30, 195,  94,
    259, 163,  62, 227, 126, 291, 146,   1, 166,  65,
    230, 129,  33, 198,  97, 262,  17, 182,  81, 246,
    150,  49, 214, 113, 278,   9, 174,  73, 238, 137,
     41, 206, 105, 270,  25, 190,  89, 254, 158,  57,
    222, 121, 286, 145,   5, 170,  69, 234, 133,  37,
    202, 101, 266,  21, 186,  85, 250, 154,  53, 218,
    117, 282,  13, 178,  77, 242, 141,  45, 210, 109,
    274,  29, 194,  93, 258, 162,  61, 226, 125, 290,
      3, 168,  67, 232, 131,  35, 200,  99, 264,  19,
    184,  83, 248, 152,  51, 216, 115, 280,  11, 176,
     75, 240, 139,  43, 208, 107, 272,  27, 192,  91,
    256, 160,  59, 224, 123, 288, 148,   7, 172,  71,
    236, 135,  39, 204, 103, 268,  23, 188,  87, 252,
    156,  55, 220, 119, 284,  15, 180,  79, 244, 143,
     47, 212, 111, 276,  31, 196,  95, 260, 164,  63,
    228, 127, 292, 
    // Size 307
      0, 179,  64, 243, 128,  32, 211,  96, 275, 144,
     16, 195,  80, 259, 163,  48, 227, 112, 291,   8,
    187,  72, 251, 136,  40, 219, 104, 283, 155,  24,
    203,  88, 267, 171,  56, 235, 120, 299,   4, 183,
     68, 247, 132,  36, 215, 100, 279, 148,  20, 199,
     84, 263, 167,  52, 231, 116, 295,  12, 191,  76,
    255, 140,  44, 223, 108, 287, 159,  28, 207,  92,
    271, 175,  60, 239, 124, 303, 152,   2, 181,  66,
    245, 130,  34, 213,  98, 277, 146,  18, 197,  82,
    261, 165,  50, 229, 114, 293,  10, 189,  74, 253,
    138,  42, 221, 106, 285, 157,  26, 205,  90, 269,
    173,  58, 237, 122, 301,   6, 185,  70, 249, 134,
     38, 217, 102, 281, 150,  22, 201,  86, 265, 169,
     54, 233, 118, 297,  14, 193,  78, 257, 142,  46,
    225, 110, 289, 161,  30, 209,  94, 273, 177,  62,
    241, 126, 305, 153,   1, 180,  65, 244, 129,  33,
    212,  97, 276, 145,  17, 196,  81, 260, 164,  49,
    228, 113, 292,   9, 188,  73, 252, 137,  41, 220,
    105, 284, 156,  25, 204,  89, 268, 172,  57, 236,
    121, 300,   5, 184,  69, 248, 133,  37, 216, 101,
    280, 149,  21, 200,  85, 264, 168,  53, 232, 117,
    296,  13, 192,  77, 256, 141,  45, 224, 109, 288,
    160,  29, 208,  93, 272, 176,  61, 240, 125, 304,
    154,   3, 182,  67, 246, 131,  35, 214,  99, 278,
    147,  19, 198,  83, 262, 166,  51, 230, 115, 294,
     11, 190,  75, 254, 139,  43, 222, 107, 286, 158,
     27, 206,  91, 270, 174,  59, 238, 123, 302,   7,
    186,  71, 250, 135,  39, 218, 103, 282, 151,  23,
    202,  87, 266, 170,  55, 234, 119, 298,  15, 194,
     79, 258, 143,  47, 226, 111, 290, 162,  31, 210,
     95, 274, 178,  63, 242, 127, 306, 
    // Size 311
      0, 183,  64, 247, 128,  32, 215,  96, 279, 144,
     16, 199,  80, 263, 167,  48, 231, 112, 295,   8,
    191,  72, 255, 136,  40, 223, 104, 287, 159,  24,
    207,  88, 271, 175,  56, 239, 120, 303, 152,   4,
    187,  68, 251, 132,  36, 219, 100, 283, 148,  20,
    203,  84, 267, 171,  52, 235, 116, 299,  12, 195,
     76, 259, 140,  44, 227, 108, 291, 163,  28, 211,
     92, 275, 179,  60, 243, 124, 307, 154,   2, 185,
     66, 249, 130,  34, 217,  98, 281, 146,  18, 201,
     82, 265, 169,  50, 233, 114, 297,  10, 193,  74,
    257, 138,  42, 225, 106, 289, 161,  26, 209,  90,
    273, 177,  58, 241, 122, 305, 157,   6, 189,  70,
    253, 134,  38, 221, 102, 285, 150,  22, 205,  86,
    269, 173,  54, 237, 118, 301,  14, 197,  78, 261,
    142,  46, 229, 110, 293, 165,  30, 213,  94, 277,
    181,  62, 245, 126, 309, 155,   1, 184,  65, 248,
    129,  33, 216,  97, 280, 145,  17, 200,  81, 264,
    168,  49, 232, 113, 296,   9, 192,  73, 256, 137,
     41, 224, 105, 288, 160,  25, 208,  89, 272, 176,
     57, 240, 121, 304, 153,   5, 188,  69, 252, 133,
     37, 220, 101, 284, 149,  21, 204,  85, 268, 172,
     53, 236, 117, 300,  13, 196,  77, 260, 141,  45,
    228, 109, 292, 164,  29, 212,  93, 276, 180,  61,
    244, 125, 308, 156,   3, 186,  67, 250, 131,  35,
    218,  99, 282, 147,  19, 202,  83, 266, 170,  51,
    234, 115, 298,  11, 194,  75, 258, 139,  43, 226,
    107, 290, 162,  27, 210,  91, 274, 178,  59, 242,
    123, 306, 158,   7, 190,  71, 254, 135,  39, 222,
    103, 286, 151,  23, 206,  87, 270, 174,  55, 238,
    119, 302,  15, 198,  79, 262, 143,  47, 230, 111,
    294, 166,  31, 214,  95, 278, 182,  63, 246, 127,
    310, 
    // Size 313
      0, 185,  64, 249, 128,  32, 217,  96, 281, 144,
     16, 201,  80, 265, 169,  48, 233, 112, 297, 152,
      8, 193,  72, 257, 136,  40, 225, 104, 289, 161,
     24, 209,  88, 273, 177,  56, 241, 120, 305,   4,
    189,  68, 253, 132,  36, 221, 100, 285, 148,  20,
    205,  84, 269, 173,  52, 237, 116, 301, 157,  12,
    197,  76, 261, 140,  44, 229, 108, 293, 165,  28,
    213,  92, 277, 181,  60, 245, 124, 309,   2, 187,
     66, 251, 130,  34, 219,  98, 283, 146,  18, 203,
     82, 267, 171,  50, 235, 114, 299, 154,  10, 195,
     74, 259, 138,  42, 227, 106, 291, 163,  26, 211,
     90, 275, 179,  58, 243, 122, 307,   6, 191,  70,
    255, 134,  38, 223, 102, 287, 150,  22, 207,  86,
    271, 175,  54, 239, 118, 303, 159,  14, 199,  78,
    263, 142,  46, 231, 110, 295, 167,  30, 215,  94,
    279, 183,  62, 247, 126, 311, 156,   1, 186,  65,
    250, 129,  33, 218,  97, 282, 145,  17, 202,  81,
    266, 170,  49, 234, 113, 298, 153,   9, 194,  73,
    258, 137,  41, 226, 105, 290, 162,  25, 210,  89,
    274, 178,  57, 242, 121, 306,   5, 190,  69, 254,
    133,  37, 222, 101, 286, 149,  21, 206,  85, 270,
    174,  53, 238, 117, 302, 158,  13, 198,  77, 262,
    141,  45, 230, 109, 294, 166,  29, 214,  93, 278,
    182,  61, 246, 125, 310,   3, 188,  67, 252, 131,
     35, 220,  99, 284, 147,  19, 204,  83, 268, 172,
     51, 236, 115, 300, 155,  11, 196,  75, 260, 139,
     43, 228, 107, 292, 164,  27, 212,  91, 276, 180,
     59, 244, 123, 308,   7, 192,  71, 256, 135,  39,
    224, 103, 288, 151,  23, 208,  87, 272, 176,  55,
    240, 119, 304, 160,  15, 200,  79, 264, 143,  47,
    232, 111, 296, 168,  31, 216,  95, 280, 184,  63,
    248, 127, 312, 
    // Size 317
      0, 189,  64, 253, 128,  32, 221,  96, 285, 144,
     16, 205,  80, 269, 173,  48, 237, 112, 301, 152,
      8, 197,  72, 261, 136,  40, 229, 104, 293, 165,
     24, 213,  88, 277, 181,  56, 245, 120, 309, 156,
      4, 193,  68, 257, 132,  36, 225, 100, 289, 148,
     20, 209,  84, 273, 177,  52, 241, 116, 305, 161,
     12, 201,  76, 265, 140,  44, 233, 108, 297, 169,
     28, 217,  92, 281, 185,  60, 249, 124, 313,   2,
    191,  66, 255, 130,  34, 223,  98, 287, 146,  18,
    207,  82, 271, 175,  50, 239, 114, 303, 154,  10,
    199,  74, 263, 138,  42, 231, 106, 295, 167,  26,
    215,  90, 279, 183,  58, 247, 122, 311, 159,   6,
    195,  70, 259, 134,  38, 227, 102, 291, 150,  22,
    211,  86, 275, 179,  54, 243, 118, 307, 163,  14,
    203,  78, 267, 142,  46, 235, 110, 299, 171,  30,
    219,  94, 283, 187,  62, 251, 126, 315, 158,   1,
    190,  65, 254, 129,  33, 222,  97, 286, 145,  17,
    206,  81, 270, 174,  49, 238, 113, 302, 153,   9,
    198,  73, 262, 137,  41, 230, 105, 294, 166,  25,
    214,  89, 278, 182,  57, 246, 121, 310, 157,   5,
    194,  69, 258, 133,  37, 226, 101, 290, 149,  21,
    210,  85, 274, 178,  53, 242, 117, 306, 162,  13,
    202,  77, 266, 141,  45, 234, 109, 298, 170,  29,
    218,  93, 282, 186,  61, 250, 125, 314,   3, 192,
     67, 256, 131,  35, 224,  99, 288, 147,  19, 208,
     83, 272, 176,  51, 240, 115, 304, 155,  11, 200,
     75, 264, 139,  43, 232, 107, 296, 168,  27, 216,
     91, 280, 184,  59, 248, 123, 312, 160,   7, 196,
     71, 260, 135,  39, 228, 103, 292, 151,  23, 212,
     87, 276, 180,  55, 244, 119, 308, 164,  15, 204,
     79, 268, 143,  47, 236, 111, 300, 172,  31, 220,
     95, 284, 188,  63, 252, 127, 316, 
    // Size 331
      0, 203, 128,  64, 267,  32, 235, 171,  96, 299,
     16, 219, 144,  80, 283,  48, 251, 187, 112, 315,
    160,   8, 211, 136,  72, 275,  40, 243, 179, 104,
    307,  24, 227, 152,  88, 291,  56, 259, 195, 120,
    323,   4, 207, 132,  68, 271,  36, 239, 175, 100,
    303,  20, 223, 148,  84, 287,  52, 255, 191, 116,
    319, 167,  12, 215, 140,  76, 279,  44, 247, 183,
    108, 311,  28, 231, 156,  92, 295,  60, 263, 199,
    124, 327, 164,   2, 205, 130,  66, 269,  34, 237,
    173,  98, 301,  18, 221, 146,  82, 285,  50, 253,
    189, 114, 317, 162,  10, 213, 138,  74, 277,  42,
    245, 181, 106, 309,  26, 229, 154,  90, 293,  58,
    261, 197, 122, 325,   6, 209, 134,  70, 273,  38,
    241, 177, 102, 305,  22, 225, 150,  86, 289,  54,
    257, 193, 118, 321, 169,  14, 217, 142,  78, 281,
     46, 249, 185, 110, 313,  30, 233, 158,  94, 297,
     62, 265, 201, 126, 329, 165,   1, 204, 129,  65,
    268,  33, 236, 172,  97, 300,  17, 220, 145,  81,
    284,  49, 252, 188, 113, 316, 161,   9, 212, 137,
     73, 276,  41, 244, 180, 105, 308,  25, 228, 153,
     89, 292,  57, 260, 196, 121, 324,   5, 208, 133,
     69, 272,  37, 240, 176, 101, 304,  21, 224, 149,
     85, 288,  53, 256, 192, 117, 320, 168,  13, 216,
    141,  77, 280,  45, 248, 184, 109, 312,  29, 232,
    157,  93, 296,  61, 264, 200, 125, 328, 166,   3,
    206, 131,  67, 270,  35, 238, 174,  99, 302,  19,
    222, 147,  83, 286,  51, 254, 190, 115, 318, 163,
     11, 214, 139,  75, 278,  43, 246, 182, 107, 310,
     27, 230, 155,  91, 294,  59, 262, 198, 123, 326,
      7, 210, 135,  71, 274,  39, 242, 178, 103, 306,
     23, 226, 151,  87, 290,  55, 258, 194, 119, 322,
    170,  15, 218, 143,  79, 282,  47, 250, 186, 111,
    314,  31, 234, 159,  95, 298,  63, 266, 202, 127,
    330, 
    // Size 337
      0, 209, 128,  64, 273,  32, 241, 177,  96, 305,
    160,  16, 225, 144,  80, 289,  48, 257, 193, 112,
    321,   8, 217, 136,  72, 281,  40, 249, 185, 104,
    313, 169,  24, 233, 152,  88, 297,  56, 265, 201,
    120, 329,   4, 213, 132,  68, 277,  36, 245, 181,
    100, 309, 164,  20, 229, 148,  84, 293,  52, 261,
    197, 116, 325,  12, 221, 140,  76, 285,  44, 253,
    189, 108, 317, 173,  28, 237, 156,  92, 301,  60,
    269, 205, 124, 333,   2, 211, 130,  66, 275,  34,
    243, 179,  98, 307, 162,  18, 227, 146,  82, 291,
     50, 259, 195, 114, 323,  10, 219, 138,  74, 283,
     42, 251, 187, 106, 315, 171,  26, 235, 154,  90,
    299,  58, 267, 203, 122, 331,   6, 215, 134,  70,
    279,  38, 247, 183, 102, 311, 166,  22, 231, 150,
     86, 295,  54, 263, 199, 118, 327,  14, 223, 142,
     78, 287,  46, 255, 191, 110, 319, 175,  30, 239,
    158,  94, 303,  62, 271, 207, 126, 335, 168,   1,
    210, 129,  65, 274,  33, 242, 178,  97, 306, 161,
     17, 226, 145,  81, 290,  49, 258, 194, 113, 322,
      9, 218, 137,  73, 282,  41, 250, 186, 105, 314,
    170,  25, 234, 153,  89, 298,  57, 266, 202, 121,
    330,   5, 214, 133,  69, 278,  37, 246, 182, 101,
    310, 165,  21, 230, 149,  85, 294,  53, 262, 198,
    117, 326,  13, 222, 141,  77, 286,  45, 254, 190,
    109, 318, 174,  29, 238, 157,  93, 302,  61, 270,
    206, 125, 334,   3, 212, 131,  67, 276,  35, 244,
    180,  99, 308, 163,  19, 228, 147,  83, 292,  51,
    260, 196, 115, 324,  11, 220, 139,  75, 284,  43,
    252, 188, 107, 316, 172,  27, 236, 155,  91, 300,
     59, 268, 204, 123, 332,   7, 216, 135,  71, 280,
     39, 248, 184, 103, 312, 167,  23, 232, 151,  87,
    296,  55, 264, 200, 119, 328,  15, 224, 143,  79,
    288,  47, 256, 192, 111, 320, 176,  31, 240, 159,
     95, 304,  63, 272, 208, 127, 336, 
    // Size 347
      0, 219, 128,  64, 283,  32, 251, 187,  96, 315,
    160,  16, 235, 144,  80, 299,  48, 267, 203, 112,
    331, 168,   8, 227, 136,  72, 291,  40, 259, 195,
    104, 323, 179,  24, 243, 152,  88, 307,  56, 275,
    211, 120, 339,   4, 223, 132,  68, 287,  36, 255,
    191, 100, 319, 164,  20, 239, 148,  84, 303,  52,
    271, 207, 116, 335, 175,  12, 231, 140,  76, 295,
     44, 263, 199, 108, 327, 183,  28, 247, 156,  92,
    311,  60, 279, 215, 124, 343, 172,   2, 221, 130,
     66, 285,  34, 253, 189,  98, 317, 162,  18, 237,
    146,  82, 301,  50, 269, 205, 114, 333, 170,  10,
    229, 138,  74, 293,  42, 261, 197, 106, 325, 181,
     26, 245, 154,  90, 309,  58, 277, 213, 122, 341,
      6, 225, 134,  70, 289,  38, 257, 193, 102, 321,
    166,  22, 241, 150,  86, 305,  54, 273, 209, 118,
    337, 177,  14, 233, 142,  78, 297,  46, 265, 201,
    110, 329, 185,  30, 249, 158,  94, 313,  62, 281,
    217, 126, 345, 173,   1, 220, 129,  65, 284,  33,
    252, 188,  97, 316, 161,  17, 236, 145,  81, 300,
     49, 268, 204, 113, 332, 169,   9, 228, 137,  73,
    292,  41, 260, 196, 105, 324, 180,  25, 244, 153,
     89, 308,  57, 276, 212, 121, 340,   5, 224, 133,
     69, 288,  37, 256, 192, 101, 320, 165,  21, 240,
    149,  85, 304,  53, 272, 208, 117, 336, 176,  13,
    232, 141,  77, 296,  45, 264, 200, 109, 328, 184,
     29, 248, 157,  93, 312,  61, 280, 216, 125, 344,
    174,   3, 222, 131,  67, 286,  35, 254, 190,  99,
    318, 163,  19, 238, 147,  83, 302,  51, 270, 206,
    115, 334, 171,  11, 230, 139,  75, 294,  43, 262,
    198, 107, 326, 182,  27, 246, 155,  91, 310,  59,
    278, 214, 123, 342,   7, 226, 135,  71, 290,  39,
    258, 194, 103, 322, 167,  23, 242, 151,  87, 306,
     55, 274, 210, 119, 338, 178,  15, 234, 143,  79,
    298,  47, 266, 202, 111, 330, 186,  31, 250, 159,
     95, 314,  63, 282, 218, 127, 346, 
    // Size 349
      0, 221, 128,  64, 285,  32, 253, 189,  96, 317,
    160,  16, 237, 144,  80, 301,  48, 269, 205, 112,
    333, 168,   8, 229, 136,  72, 293,  40, 261, 197,
    104, 325, 181,  24, 245, 152,  88, 309,  56, 277,
    213, 120, 341, 172,   4, 225, 132,  68, 289,  36,
    257, 193, 100, 321, 164,  20, 241, 148,  84, 305,
     52, 273, 209, 116, 337, 177,  12, 233, 140,  76,
    297,  44, 265, 201, 108, 329, 185,  28, 249, 156,
     92, 313,  60, 281, 217, 124, 345,   2, 223, 130,
     66, 287,  34, 255, 191,  98, 319, 162,  18, 239,
    146,  82, 303,  50, 271, 207, 114, 335, 170,  10,
    231, 138,  74, 295,  42, 263, 199, 106, 327, 183,
     26, 247, 154,  90, 311,  58, 279, 215, 122, 343,
    175,   6, 227, 134,  70, 291,  38, 259, 195, 102,
    323, 166,  22, 243, 150,  86, 307,  54, 275, 211,
    118, 339, 179,  14, 235, 142,  78, 299,  46, 267,
    203, 110, 331, 187,  30, 251, 158,  94, 315,  62,
    283, 219, 126, 347, 174,   1, 222, 129,  65, 286,
     33, 254, 190,  97, 318, 161,  17, 238, 145,  81,
    302,  49, 270, 206, 113, 334, 169,   9, 230, 137,
     73, 294,  41, 262, 198, 105, 326, 182,  25, 246,
    153,  89, 310,  57, 278, 214, 121, 342, 173,   5,
    226, 133,  69, 290,  37, 258, 194, 101, 322, 165,
     21, 242, 149,  85, 306,  53, 274, 210, 117, 338,
    178,  13, 234, 141,  77, 298,  45, 266, 202, 109,
    330, 186,  29, 250, 157,  93, 314,  61, 282, 218,
    125, 346,   3, 224, 131,  67, 288,  35, 256, 192,
     99, 320, 163,  19, 240, 147,  83, 304,  51, 272,
    208, 115, 336, 171,  11, 232, 139,  75, 296,  43,
    264, 200, 107, 328, 184,  27, 248, 155,  91, 312,
     59, 280, 216, 123, 344, 176,   7, 228, 135,  71,
    292,  39, 260, 196, 103, 324, 167,  23, 244, 151,
     87, 308,  55, 276, 212, 119, 340, 180,  15, 236,
    143,  79, 300,  47, 268, 204, 111, 332, 188,  31,
    252, 159,  95, 316,  63, 284, 220, 127, 348, 
    // Size 353
      0, 225, 128,  64, 289, 160,  32, 257, 193,  96,
    321,  16, 241, 144,  80, 305, 177,  48, 273, 209,
    112, 337,   8, 233, 136,  72, 297, 168,  40, 265,
    201, 104, 329,  24, 249, 152,  88, 313, 185,  56,
    281, 217, 120, 345,   4, 229, 132,  68, 293, 164,
     36, 261, 197, 100, 325,  20, 245, 148,  84, 309,
    181,  52, 277, 213, 116, 341,  12, 237, 140,  76,
    301, 172,  44, 269, 205, 108, 333,  28, 253, 156,
     92, 317, 189,  60, 285, 221, 124, 349,   2, 227,
    130,  66, 291, 162,  34, 259, 195,  98, 323,  18,
    243, 146,  82, 307, 179,  50, 275, 211, 114, 339,
     10, 235, 138,  74, 299, 170,  42, 267, 203, 106,
    331,  26, 251, 154,  90, 315, 187,  58, 283, 219,
    122, 347,   6, 231, 134,  70, 295, 166,  38, 263,
    199, 102, 327,  22, 247, 150,  86, 311, 183,  54,
    279, 215, 118, 343,  14, 239, 142,  78, 303, 174,
     46, 271, 207, 110, 335,  30, 255, 158,  94, 319,
    191,  62, 287, 223, 126, 351, 176,   1, 226, 129,
     65, 290, 161,  33, 258, 194,  97, 322,  17, 242,
    145,  81, 306, 178,  49, 274, 210, 113, 338,   9,
    234, 137,  73, 298, 169,  41, 266, 202, 105, 330,
     25, 250, 153,  89, 314, 186,  57, 282, 218, 121,
    346,   5, 230, 133,  69, 294, 165,  37, 262, 198,
    101, 326,  21, 246, 149,  85, 310, 182,  53, 278,
    214, 117, 342,  13, 238, 141,  77, 302, 173,  45,
    270, 206, 109, 334,  29, 254, 157,  93, 318, 190,
     61, 286, 222, 125, 350,   3, 228, 131,  67, 292,
    163,  35, 260, 196,  99, 324,  19, 244, 147,  83,
    308, 180,  51, 276, 212, 115, 340,  11, 236, 139,
     75, 300, 171,  43, 268, 204, 107, 332,  27, 252,
    155,  91, 316, 188,  59, 284, 220, 123, 348,   7,
    232, 135,  71, 296, 167,  39, 264, 200, 103, 328,
     23, 248, 151,  87, 312, 184,  55, 280, 216, 119,
    344,  15, 240, 143,  79, 304, 175,  47, 272, 208,
    111, 336,  31, 256, 159,  95, 320, 192,  63, 288,
    224, 127, 352, 
    // Size 359
      0, 231, 128,  64, 295, 160,  32, 263, 199,  96,
    327,  16, 247, 144,  80, 311, 183,  48, 279, 215,
    112, 343,   8, 239, 136,  72, 303, 168,  40, 271,
    207, 104, 335,  24, 255, 152,  88, 319, 191,  56,
    287, 223, 120, 351, 176,   4, 235, 132,  68, 299,
    164,  36, 267, 203, 100, 331,  20, 251, 148,  84,
    315, 187,  52, 283, 219, 116, 347,  12, 243, 140,
     76, 307, 172,  44, 275, 211, 108, 339,  28, 259,
    156,  92, 323, 195,  60, 291, 227, 124, 355, 178,
      2, 233, 130,  66, 297, 162,  34, 265, 201,  98,
    329,  18, 249, 146,  82, 313, 185,  50, 281, 217,
    114, 345,  10, 241, 138,  74, 305, 170,  42, 273,
    209, 106, 337,  26, 257, 154,  90, 321, 193,  58,
    289, 225, 122, 353, 181,   6, 237, 134,  70, 301,
    166,  38, 269, 205, 102, 333,  22, 253, 150,  86,
    317, 189,  54, 285, 221, 118, 349,  14, 245, 142,
     78, 309, 174,  46, 277, 213, 110, 341,  30, 261,
    158,  94, 325, 197,  62, 293, 229, 126, 357, 179,
      1, 232, 129,  65, 296, 161,  33, 264, 200,  97,
    328,  17, 248, 145,  81, 312, 184,  49, 280, 216,
    113, 344,   9, 240, 137,  73, 304, 169,  41, 272,
    208, 105, 336,  25, 256, 153,  89, 320, 192,  57,
    288, 224, 121, 352, 177,   5, 236, 133,  69, 300,
    165,  37, 268, 204, 101, 332,  21, 252, 149,  85,
    316, 188,  53, 284, 220, 117, 348,  13, 244, 141,
     77, 308, 173,  45, 276, 212, 109, 340,  29, 260,
    157,  93, 324, 196,  61, 292, 228, 125, 356, 180,
      3, 234, 131,  67, 298, 163,  35, 266, 202,  99,
    330,  19, 250, 147,  83, 314, 186,  51, 282, 218,
    115, 346,  11, 242, 139,  75, 306, 171,  43, 274,
    210, 107, 338,  27, 258, 155,  91, 322, 194,  59,
    290, 226, 123, 354, 182,   7, 238, 135,  71, 302,
    167,  39, 270, 206, 103, 334,  23, 254, 151,  87,
    318, 190,  55, 286, 222, 119, 350,  15, 246, 143,
     79, 310, 175,  47, 278, 214, 111, 342,  31, 262,
    159,  95, 326, 198,  63, 294, 230, 127, 358, 
    // Size 367
      0, 239, 128,  64, 303, 160,  32, 271, 207,  96,
    335,  16, 255, 144,  80, 319, 191,  48, 287, 223,
    112, 351, 176,   8, 247, 136,  72, 311, 168,  40,
    279, 215, 104, 343,  24, 263, 152,  88, 327, 199,
     56, 295, 231, 120, 359, 180,   4, 243, 132,  68,
    307, 164,  36, 275, 211, 100, 339,  20, 259, 148,
     84, 323, 195,  52, 291, 227, 116, 355, 187,  12,
    251, 140,  76, 315, 172,  44, 283, 219, 108, 347,
     28, 267, 156,  92, 331, 203,  60, 299, 235, 124,
    363, 182,   2, 241, 130,  66, 305, 162,  34, 273,
    209,  98, 337,  18, 257, 146,  82, 321, 193,  50,
    289, 225, 114, 353, 178,  10, 249, 138,  74, 313,
    170,  42, 281, 217, 106, 345,  26, 265, 154,  90,
    329, 201,  58, 297, 233, 122, 361, 185,   6, 245,
    134,  70, 309, 166,  38, 277, 213, 102, 341,  22,
    261, 150,  86, 325, 197,  54, 293, 229, 118, 357,
    189,  14, 253, 142,  78, 317, 174,  46, 285, 221,
    110, 349,  30, 269, 158,  94, 333, 205,  62, 301,
    237, 126, 365, 183,   1, 240, 129,  65, 304, 161,
     33, 272, 208,  97, 336,  17, 256, 145,  81, 320,
    192,  49, 288, 224, 113, 352, 177,   9, 248, 137,
     73, 312, 169,  41, 280, 216, 105, 344,  25, 264,
    153,  89, 328, 200,  57, 296, 232, 121, 360, 181,
      5, 244, 133,  69, 308, 165,  37, 276, 212, 101,
    340,  21, 260, 149,  85, 324, 196,  53, 292, 228,
    117, 356, 188,  13, 252, 141,  77, 316, 173,  45,
    284, 220, 109, 348,  29, 268, 157,  93, 332, 204,
     61, 300, 236, 125, 364, 184,   3, 242, 131,  67,
    306, 163,  35, 274, 210,  99, 338,  19, 258, 147,
     83, 322, 194,  51, 290, 226, 115, 354, 179,  11,
    250, 139,  75, 314, 171,  43, 282, 218, 107, 346,
     27, 266, 155,  91, 330, 202,  59, 298, 234, 123,
    362, 186,   7, 246, 135,  71, 310, 167,  39, 278,
    214, 103, 342,  23, 262, 151,  87, 326, 198,  55,
    294, 230, 119, 358, 190,  15, 254, 143,  79, 318,
    175,  47, 286, 222, 111, 350,  31, 270, 159,  95,
    334, 206,  63, 302, 238, 127, 366, 
    // Size 373
      0, 245, 128,  64, 309, 160,  32, 277, 213,  96,
    341, 176,  16, 261, 144,  80, 325, 197,  48, 293,
    229, 112, 357,   8, 253, 136,  72, 317, 168,  40,
    285, 221, 104, 349, 189,  24, 269, 152,  88, 333,
    205,  56, 301, 237, 120, 365, 184,   4, 249, 132,
     68, 313, 164,  36, 281, 217, 100, 345, 180,  20,
    265, 148,  84, 329, 201,  52, 297, 233, 116, 361,
     12, 257, 140,  76, 321, 172,  44, 289, 225, 108,
    353, 193,  28, 273, 156,  92, 337, 209,  60, 305,
    241, 124, 369,   2, 247, 130,  66, 311, 162,  34,
    279, 215,  98, 343, 178,  18, 263, 146,  82, 327,
    199,  50, 295, 231, 114, 359,  10, 255, 138,  74,
    319, 170,  42, 287, 223, 106, 351, 191,  26, 271,
    154,  90, 335, 207,  58, 303, 239, 122, 367, 187,
      6, 251, 134,  70, 315, 166,  38, 283, 219, 102,
    347, 182,  22, 267, 150,  86, 331, 203,  54, 299,
    235, 118, 363,  14, 259, 142,  78, 323, 174,  46,
    291, 227, 110, 355, 195,  30, 275, 158,  94, 339,
    211,  62, 307, 243, 126, 371, 186,   1, 246, 129,
     65, 310, 161,  33, 278, 214,  97, 342, 177,  17,
    262, 145,  81, 326, 198,  49, 294, 230, 113, 358,
      9, 254, 137,  73, 318, 169,  41, 286, 222, 105,
    350, 190,  25, 270, 153,  89, 334, 206,  57, 302,
    238, 121, 366, 185,   5, 250, 133,  69, 314, 165,
     37, 282, 218, 101, 346, 181,  21, 266, 149,  85,
    330, 202,  53, 298, 234, 117, 362,  13, 258, 141,
     77, 322, 173,  45, 290, 226, 109, 354, 194,  29,
    274, 157,  93, 338, 210,  61, 306, 242, 125, 370,
      3, 248, 131,  67, 312, 163,  35, 280, 216,  99,
    344, 179,  19, 264, 147,  83, 328, 200,  51, 296,
    232, 115, 360,  11, 256, 139,  75, 320, 171,  43,
    288, 224, 107, 352, 192,  27, 272, 155,  91, 336,
    208,  59, 304, 240, 123, 368, 188,   7, 252, 135,
     71, 316, 167,  39, 284, 220, 103, 348, 183,  23,
    268, 151,  87, 332, 204,  55, 300, 236, 119, 364,
     15, 260, 143,  79, 324, 175,  47, 292, 228, 111,
    356, 196,  31, 276, 159,  95, 340, 212,  63, 308,
    244, 127, 372, 
    // Size 379
      0, 251, 128,  64, 315, 160,  32, 283, 219,  96,
    347, 176,  16, 267, 144,  80, 331, 203,  48, 299,
    235, 112, 363, 184,   8, 259, 136,  72, 323, 168,
     40, 291, 227, 104, 355, 195,  24, 275, 152,  88,
    339, 211,  56, 307, 243, 120, 371,   4, 255, 132,
     68, 319, 164,  36, 287, 223, 100, 351, 180,  20,
    271, 148,  84, 335, 207,  52, 303, 239, 116, 367,
    191,  12, 263, 140,  76, 327, 172,  44, 295, 231,
    108, 359, 199,  28, 279, 156,  92, 343, 215,  60,
    311, 247, 124, 375, 188,   2, 253, 130,  66, 317,
    162,  34, 285, 221,  98, 349, 178,  18, 269, 146,
     82, 333, 205,  50, 301, 237, 114, 365, 186,  10,
    261, 138,  74, 325, 170,  42, 293, 229, 106, 357,
    197,  26, 277, 154,  90, 341, 213,  58, 309, 245,
    122, 373,   6, 257, 134,  70, 321, 166,  38, 289,
    225, 102, 353, 182,  22, 273, 150,  86, 337, 209,
     54, 305, 241, 118, 369, 193,  14, 265, 142,  78,
    329, 174,  46, 297, 233, 110, 361, 201,  30, 281,
    158,  94, 345, 217,  62, 313, 249, 126, 377, 189,
      1, 252, 129,  65, 316, 161,  33, 284, 220,  97,
    348, 177,  17, 268, 145,  81, 332, 204,  49, 300,
    236, 113, 364, 185,   9, 260, 137,  73, 324, 169,
     41, 292, 228, 105, 356, 196,  25, 276, 153,  89,
    340, 212,  57, 308, 244, 121, 372,   5, 256, 133,
     69, 320, 165,  37, 288, 224, 101, 352, 181,  21,
    272, 149,  85, 336, 208,  53, 304, 240, 117, 368,
    192,  13, 264, 141,  77, 328, 173,  45, 296, 232,
    109, 360, 200,  29, 280, 157,  93, 344, 216,  61,
    312, 248, 125, 376, 190,   3, 254, 131,  67, 318,
    163,  35, 286, 222,  99, 350, 179,  19, 270, 147,
     83, 334, 206,  51, 302, 238, 115, 366, 187,  11,
    262, 139,  75, 326, 171,  43, 294, 230, 107, 358,
    198,  27, 278, 155,  91, 342, 214,  59, 310, 246,
    123, 374,   7, 258, 135,  71, 322, 167,  39, 290,
    226, 103, 354, 183,  23, 274, 151,  87, 338, 210,
     55, 306, 242, 119, 370, 194,  15, 266, 143,  79,
    330, 175,  47, 298, 234, 111, 362, 202,  31, 282,
    159,  95, 346, 218,  63, 314, 250, 127, 378, 
    // Size 383
      0, 255, 128,  64, 319, 160,  32, 287, 223,  96,
    351, 176,  16, 271, 144,  80, 335, 207,  48, 303,
    239, 112, 367, 184,   8, 263, 136,  72, 327, 168,
     40, 295, 231, 104, 359, 199,  24, 279, 152,  88,
    343, 215,  56, 311, 247, 120, 375, 188,   4, 259,
    132,  68, 323, 164,  36, 291, 227, 100, 355, 180,
     20, 275, 148,  84, 339, 211,  52, 307, 243, 116,
    371, 195,  12, 267, 140,  76, 331, 172,  44, 299,
    235, 108, 363, 203,  28, 283, 156,  92, 347, 219,
     60, 315, 251, 124, 379, 190,   2, 257, 130,  66,
    321, 162,  34, 289, 225,  98, 353, 178,  18, 273,
    146,  82, 337, 209,  50, 305, 241, 114, 369, 186,
     10, 265, 138,  74, 329, 170,  42, 297, 233, 106,
    361, 201,  26, 281, 154,  90, 345, 217,  58, 313,
    249, 122, 377, 193,   6, 261, 134,  70, 325, 166,
     38, 293, 229, 102, 357, 182,  22, 277, 150,  86,
    341, 213,  54, 309, 245, 118, 373, 197,  14, 269,
    142,  78, 333, 174,  46, 301, 237, 110, 365, 205,
     30, 285, 158,  94, 349, 221,  62, 317, 253, 126,
    381, 191,   1, 256, 129,  65, 320, 161,  33, 288,
    224,  97, 352, 177,  17, 272, 145,  81, 336, 208,
     49, 304, 240, 113, 368, 185,   9, 264, 137,  73,
    328, 169,  41, 296, 232, 105, 360, 200,  25, 280,
    153,  89, 344, 216,  57, 312, 248, 121, 376, 189,
      5, 260, 133,  69, 324, 165,  37, 292, 228, 101,
    356, 181,  21, 276, 149,  85, 340, 212,  53, 308,
    244, 117, 372, 196,  13, 268, 141,  77, 332, 173,
     45, 300, 236, 109, 364, 204,  29, 284, 157,  93,
    348, 220,  61, 316, 252, 125, 380, 192,   3, 258,
    131,  67, 322, 163,  35, 290, 226,  99, 354, 179,
     19, 274, 147,  83, 338, 210,  51, 306, 242, 115,
    370, 187,  11, 266, 139,  75, 330, 171,  43, 298,
    234, 107, 362, 202,  27, 282, 155,  91, 346, 218,
     59, 314, 250, 123, 378, 194,   7, 262, 135,  71,
    326, 167,  39, 294, 230, 103, 358, 183,  23, 278,
    151,  87, 342, 214,  55, 310, 246, 119, 374, 198,
     15, 270, 143,  79, 334, 175,  47, 302, 238, 111,
    366, 206,  31, 286, 159,  95, 350, 222,  63, 318,
    254, 127, 382, 
    // Size 389
      0, 128, 261,  64, 197, 325,  32, 160, 293,  96,
    229, 357,  16, 144, 277,  80, 213, 341,  48, 176,
    309, 112, 245, 373,   8, 136, 269,  72, 205, 333,
     40, 168, 301, 104, 237, 365,  24, 152, 285,  88,
    221, 349,  56, 184, 317, 120, 253, 381, 192,   4,
    132, 265,  68, 201, 329,  36, 164, 297, 100, 233,
    361,  20, 148, 281,  84, 217, 345,  52, 180, 313,
    116, 249, 377,  12, 140, 273,  76, 209, 337,  44,
    172, 305, 108, 241, 369,  28, 156, 289,  92, 225,
    353,  60, 188, 321, 124, 257, 385,   2, 130, 263,
     66, 199, 327,  34, 162, 295,  98, 231, 359,  18,
    146, 279,  82, 215, 343,  50, 178, 311, 114, 247,
    375,  10, 138, 271,  74, 207, 335,  42, 170, 303,
    106, 239, 367,  26, 154, 287,  90, 223, 351,  58,
    186, 319, 122, 255, 383, 195,   6, 134, 267,  70,
    203, 331,  38, 166, 299, 102, 235, 363,  22, 150,
    283,  86, 219, 347,  54, 182, 315, 118, 251, 379,
     14, 142, 275,  78, 211, 339,  46, 174, 307, 110,
    243, 371,  30, 158, 291,  94, 227, 355,  62, 190,
    323, 126, 259, 387, 194,   1, 129, 262,  65, 198,
    326,  33, 161, 294,  97, 230, 358,  17, 145, 278,
     81, 214, 342,  49, 177, 310, 113, 246, 374,   9,
    137, 270,  73, 206, 334,  41, 169, 302, 105, 238,
    366,  25, 153, 286,  89, 222, 350,  57, 185, 318,
    121, 254, 382, 193,   5, 133, 266,  69, 202, 330,
     37, 165, 298, 101, 234, 362,  21, 149, 282,  85,
    218, 346,  53, 181, 314, 117, 250, 378,  13, 141,
    274,  77, 210, 338,  45, 173, 306, 109, 242, 370,
     29, 157, 290,  93, 226, 354,  61, 189, 322, 125,
    258, 386,   3, 131, 264,  67, 200, 328,  35, 163,
    296,  99, 232, 360,  19, 147, 280,  83, 216, 344,
     51, 179, 312, 115, 248, 376,  11, 139, 272,  75,
    208, 336,  43, 171, 304, 107, 240, 368,  27, 155,
    288,  91, 224, 352,  59, 187, 320, 123, 256, 384,
    196,   7, 135, 268,  71, 204, 332,  39, 167, 300,
    103, 236, 364,  23, 151, 284,  87, 220, 348,  55,
    183, 316, 119, 252, 380,  15, 143, 276,  79, 212,
    340,  47, 175, 308, 111, 244, 372,  31, 159, 292,
     95, 228, 356,  63, 191, 324, 127, 260, 388, 
    // Size 397
      0, 128, 269,  64, 205, 333,  32, 160, 301,  96,
    237, 365,  16, 144, 285,  80, 221, 349,  48, 176,
    317, 112, 253, 381, 192,   8, 136, 277,  72, 213,
    341,  40, 168, 309, 104, 245, 373,  24, 152, 293,
     88, 229, 357,  56, 184, 325, 120, 261, 389, 196,
      4, 132, 273,  68, 209, 337,  36, 164, 305, 100,
    241, 369,  20, 148, 289,  84, 225, 353,  52, 180,
    321, 116, 257, 385, 201,  12, 140, 281,  76, 217,
    345,  44, 172, 313, 108, 249, 377,  28, 156, 297,
     92, 233, 361,  60, 188, 329, 124, 265, 393,   2,
    130, 271,  66, 207, 335,  34, 162, 303,  98, 239,
    367,  18, 146, 287,  82, 223, 351,  50, 178, 319,
    114, 255, 383, 194,  10, 138, 279,  74, 215, 343,
     42, 170, 311, 106, 247, 375,  26, 154, 295,  90,
    231, 359,  58, 186, 327, 122, 263, 391, 199,   6,
    134, 275,  70, 211, 339,  38, 166, 307, 102, 243,
    371,  22, 150, 291,  86, 227, 355,  54, 182, 323,
    118, 259, 387, 203,  14, 142, 283,  78, 219, 347,
     46, 174, 315, 110, 251, 379,  30, 158, 299,  94,
    235, 363,  62, 190, 331, 126, 267, 395, 198,   1,
    129, 270,  65, 206, 334,  33, 161, 302,  97, 238,
    366,  17, 145, 286,  81, 222, 350,  49, 177, 318,
    113, 254, 382, 193,   9, 137, 278,  73, 214, 342,
     41, 169, 310, 105, 246, 374,  25, 153, 294,  89,
    230, 358,  57, 185, 326, 121, 262, 390, 197,   5,
    133, 274,  69, 210, 338,  37, 165, 306, 101, 242,
    370,  21, 149, 290,  85, 226, 354,  53, 181, 322,
    117, 258, 386, 202,  13, 141, 282,  77, 218, 346,
     45, 173, 314, 109, 250, 378,  29, 157, 298,  93,
    234, 362,  61, 189, 330, 125, 266, 394,   3, 131,
    272,  67, 208, 336,  35, 163, 304,  99, 240, 368,
     19, 147, 288,  83, 224, 352,  51, 179, 320, 115,
    256, 384, 195,  11, 139, 280,  75, 216, 344,  43,
    171, 312, 107, 248, 376,  27, 155, 296,  91, 232,
    360,  59, 187, 328, 123, 264, 392, 200,   7, 135,
    276,  71, 212, 340,  39, 167, 308, 103, 244, 372,
     23, 151, 292,  87, 228, 356,  55, 183, 324, 119,
    260, 388, 204,  15, 143, 284,  79, 220, 348,  47,
    175, 316, 111, 252, 380,  31, 159, 300,  95, 236,
    364,  63, 191, 332, 127, 268, 396, 
    // Size 401
      0, 128, 273,  64, 209, 337,  32, 160, 305,  96,
    241, 369, 192,  16, 144, 289,  80, 225, 353,  48,
    176, 321, 112, 257, 385,   8, 136, 281,  72, 217,
    345,  40, 168, 313, 104, 249, 377, 201,  24, 152,
    297,  88, 233, 361,  56, 184, 329, 120, 265, 393,
      4, 132, 277,  68, 213, 341,  36, 164, 309, 100,
    245, 373, 196,  20, 148, 293,  84, 229, 357,  52,
    180, 325, 116, 261, 389,  12, 140, 285,  76, 221,
    349,  44, 172, 317, 108, 253, 381, 205,  28, 156,
    301,  92, 237, 365,  60, 188, 333, 124, 269, 397,
      2, 130, 275,  66, 211, 339,  34, 162, 307,  98,
    243, 371, 194,  18, 146, 291,  82, 227, 355,  50,
    178, 323, 114, 259, 387,  10, 138, 283,  74, 219,
    347,  42, 170, 315, 106, 251, 379, 203,  26, 154,
    299,  90, 235, 363,  58, 186, 331, 122, 267, 395,
      6, 134, 279,  70, 215, 343,  38, 166, 311, 102,
    247, 375, 198,  22, 150, 295,  86, 231, 359,  54,
    182, 327, 118, 263, 391,  14, 142, 287,  78, 223,
    351,  46, 174, 319, 110, 255, 383, 207,  30, 158,
    303,  94, 239, 367,  62, 190, 335, 126, 271, 399,
    200,   1, 129, 274,  65, 210, 338,  33, 161, 306,
     97, 242, 370, 193,  17, 145, 290,  81, 226, 354,
     49, 177, 322, 113, 258, 386,   9, 137, 282,  73,
    218, 346,  41, 169, 314, 105, 250, 378, 202,  25,
    153, 298,  89, 234, 362,  57, 185, 330, 121, 266,
    394,   5, 133, 278,  69, 214, 342,  37, 165, 310,
    101, 246, 374, 197,  21, 149, 294,  85, 230, 358,
     53, 181, 326, 117, 262, 390,  13, 141, 286,  77,
    222, 350,  45, 173, 318, 109, 254, 382, 206,  29,
    157, 302,  93, 238, 366,  61, 189, 334, 125, 270,
    398,   3, 131, 276,  67, 212, 340,  35, 163, 308,
     99, 244, 372, 195,  19, 147, 292,  83, 228, 356,
     51, 179, 324, 115, 260, 388,  11, 139, 284,  75,
    220, 348,  43, 171, 316, 107, 252, 380, 204,  27,
    155, 300,  91, 236, 364,  59, 187, 332, 123, 268,
    396,   7, 135, 280,  71, 216, 344,  39, 167, 312,
    103, 248, 376, 199,  23, 151, 296,  87, 232, 360,
     55, 183, 328, 119, 264, 392,  15, 143, 288,  79,
    224, 352,  47, 175, 320, 111, 256, 384, 208,  31,
    159, 304,  95, 240, 368,  63, 191, 336, 127, 272,
    400, 
    // Size 409
      0, 128, 281,  64, 217, 345,  32, 160, 313,  96,
    249, 377, 192,  16, 144, 297,  80, 233, 361,  48,
    176, 329, 112, 265, 393, 200,   8, 136, 289,  72,
    225, 353,  40, 168, 321, 104, 257, 385, 209,  24,
    152, 305,  88, 241, 369,  56, 184, 337, 120, 273,
    401,   4, 132, 285,  68, 221, 349,  36, 164, 317,
    100, 253, 381, 196,  20, 148, 301,  84, 237, 365,
     52, 180, 333, 116, 269, 397, 205,  12, 140, 293,
     76, 229, 357,  44, 172, 325, 108, 261, 389, 213,
     28, 156, 309,  92, 245, 373,  60, 188, 341, 124,
    277, 405,   2, 130, 283,  66, 219, 347,  34, 162,
    315,  98, 251, 379, 194,  18, 146, 299,  82, 235,
    363,  50, 178, 331, 114, 267, 395, 202,  10, 138,
    291,  74, 227, 355,  42, 170, 323, 106, 259, 387,
    211,  26, 154, 307,  90, 243, 371,  58, 186, 339,
    122, 275, 403,   6, 134, 287,  70, 223, 351,  38,
    166, 319, 102, 255, 383, 198,  22, 150, 303,  86,
    239, 367,  54, 182, 335, 118, 271, 399, 207,  14,
    142, 295,  78, 231, 359,  46, 174, 327, 110, 263,
    391, 215,  30, 158, 311,  94, 247, 375,  62, 190,
    343, 126, 279, 407, 204,   1, 129, 282,  65, 218,
    346,  33, 161, 314,  97, 250, 378, 193,  17, 145,
    298,  81, 234, 362,  49, 177, 330, 113, 266, 394,
    201,   9, 137, 290,  73, 226, 354,  41, 169, 322,
    105, 258, 386, 210,  25, 153, 306,  89, 242, 370,
     57, 185, 338, 121, 274, 402,   5, 133, 286,  69,
    222, 350,  37, 165, 318, 101, 254, 382, 197,  21,
    149, 302,  85, 238, 366,  53, 181, 334, 117, 270,
    398, 206,  13, 141, 294,  77, 230, 358,  45, 173,
    326, 109, 262, 390, 214,  29, 157, 310,  93, 246,
    374,  61, 189, 342, 125, 278, 406,   3, 131, 284,
     67, 220, 348,  35, 163, 316,  99, 252, 380, 195,
     19, 147, 300,  83, 236, 364,  51, 179, 332, 115,
    268, 396, 203,  11, 139, 292,  75, 228, 356,  43,
    171, 324, 107, 260, 388, 212,  27, 155, 308,  91,
    244, 372,  59, 187, 340, 123, 276, 404,   7, 135,
    288,  71, 224, 352,  39, 167, 320, 103, 256, 384,
    199,  23, 151, 304,  87, 240, 368,  55, 183, 336,
    119, 272, 400, 208,  15, 143, 296,  79, 232, 360,
     47, 175, 328, 111, 264, 392, 216,  31, 159, 312,
     95, 248, 376,  63, 191, 344, 127, 280, 408, 
    // Size 419
      0, 128, 291,  64, 227, 355, 192,  32, 160, 323,
     96, 259, 387,  16, 144, 307,  80, 243, 371, 211,
     48, 176, 339, 112, 275, 403,   8, 136, 299,  72,
    235, 363, 200,  40, 168, 331, 104, 267, 395,  24,
    152, 315,  88, 251, 379, 219,  56, 184, 347, 120,
    283, 411,   4, 132, 295,  68, 231, 359, 196,  36,
    164, 327, 100, 263, 391,  20, 148, 311,  84, 247,
    375, 215,  52, 180, 343, 116, 279, 407,  12, 140,
    303,  76, 239, 367, 204,  44, 172, 335, 108, 271,
    399,  28, 156, 319,  92, 255, 383, 223,  60, 188,
    351, 124, 287, 415, 208,   2, 130, 293,  66, 229,
    357, 194,  34, 162, 325,  98, 261, 389,  18, 146,
    309,  82, 245, 373, 213,  50, 178, 341, 114, 277,
    405,  10, 138, 301,  74, 237, 365, 202,  42, 170,
    333, 106, 269, 397,  26, 154, 317,  90, 253, 381,
    221,  58, 186, 349, 122, 285, 413,   6, 134, 297,
     70, 233, 361, 198,  38, 166, 329, 102, 265, 393,
     22, 150, 313,  86, 249, 377, 217,  54, 182, 345,
    118, 281, 409,  14, 142, 305,  78, 241, 369, 206,
     46, 174, 337, 110, 273, 401,  30, 158, 321,  94,
    257, 385, 225,  62, 190, 353, 126, 289, 417, 209,
      1, 129, 292,  65, 228, 356, 193,  33, 161, 324,
     97, 260, 388,  17, 145, 308,  81, 244, 372, 212,
     49, 177, 340, 113, 276, 404,   9, 137, 300,  73,
    236, 364, 201,  41, 169, 332, 105, 268, 396,  25,
    153, 316,  89, 252, 380, 220,  57, 185, 348, 121,
    284, 412,   5, 133, 296,  69, 232, 360, 197,  37,
    165, 328, 101, 264, 392,  21, 149, 312,  85, 248,
    376, 216,  53, 181, 344, 117, 280, 408,  13, 141,
    304,  77, 240, 368, 205,  45, 173, 336, 109, 272,
    400,  29, 157, 320,  93, 256, 384, 224,  61, 189,
    352, 125, 288, 416, 210,   3, 131, 294,  67, 230,
    358, 195,  35, 163, 326,  99, 262, 390,  19, 147,
    310,  83, 246, 374, 214,  51, 179, 342, 115, 278,
    406,  11, 139, 302,  75, 238, 366, 203,  43, 171,
    334, 107, 270, 398,  27, 155, 318,  91, 254, 382,
    222,  59, 187, 350, 123, 286, 414,   7, 135, 298,
     71, 234, 362, 199,  39, 167, 330, 103, 266, 394,
     23, 151, 314,  87, 250, 378, 218,  55, 183, 346,
    119, 282, 410,  15, 143, 306,  79, 242, 370, 207,
     47, 175, 338, 111, 274, 402,  31, 159, 322,  95,
    258, 386, 226,  63, 191, 354, 127, 290, 418, 
    // Size 421
      0, 128, 293,  64, 229, 357, 192,  32, 160, 325,
     96, 261, 389,  16, 144, 309,  80, 245, 373, 213,
     48, 176, 341, 112, 277, 405,   8, 136, 301,  72,
    237, 365, 200,  40, 168, 333, 104, 269, 397,  24,
    152, 317,  88, 253, 381, 221,  56, 184, 349, 120,
    285, 413, 208,   4, 132, 297,  68, 233, 361, 196,
     36, 164, 329, 100, 265, 393,  20, 148, 313,  84,
    249, 377, 217,  52, 180, 345, 116, 281, 409,  12,
    140, 305,  76, 241, 369, 204,  44, 172, 337, 108,
    273, 401,  28, 156, 321,  92, 257, 385, 225,  60,
    188, 353, 124, 289, 417,   2, 130, 295,  66, 231,
    359, 194,  34, 162, 327,  98, 263, 391,  18, 146,
    311,  82, 247, 375, 215,  50, 178, 343, 114, 279,
    407,  10, 138, 303,  74, 239, 367, 202,  42, 170,
    335, 106, 271, 399,  26, 154, 319,  90, 255, 383,
    223,  58, 186, 351, 122, 287, 415, 211,   6, 134,
    299,  70, 235, 363, 198,  38, 166, 331, 102, 267,
    395,  22, 150, 315,  86, 251, 379, 219,  54, 182,
    347, 118, 283, 411,  14, 142, 307,  78, 243, 371,
    206,  46, 174, 339, 110, 275, 403,  30, 158, 323,
     94, 259, 387, 227,  62, 190, 355, 126, 291, 419,
    210,   1, 129, 294,  65, 230, 358, 193,  33, 161,
    326,  97, 262, 390,  17, 145, 310,  81, 246, 374,
    214,  49, 177, 342, 113, 278, 406,   9, 137, 302,
     73, 238, 366, 201,  41, 169, 334, 105, 270, 398,
     25, 153, 318,  89, 254, 382, 222,  57, 185, 350,
    121, 286, 414, 209,   5, 133, 298,  69, 234, 362,
    197,  37, 165, 330, 101, 266, 394,  21, 149, 314,
     85, 250, 378, 218,  53, 181, 346, 117, 282, 410,
     13, 141, 306,  77, 242, 370, 205,  45, 173, 338,
    109, 274, 402,  29, 157, 322,  93, 258, 386, 226,
     61, 189, 354, 125, 290, 418,   3, 131, 296,  67,
    232, 360, 195,  35, 163, 328,  99, 264, 392,  19,
    147, 312,  83, 248, 376, 216,  51, 179, 344, 115,
    280, 408,  11, 139, 304,  75, 240, 368, 203,  43,
    171, 336, 107, 272, 400,  27, 155, 320,  91, 256,
    384, 224,  59, 187, 352, 123, 288, 416, 212,   7,
    135, 300,  71, 236, 364, 199,  39, 167, 332, 103,
    268, 396,  23, 151, 316,  87, 252, 380, 220,  55,
    183, 348, 119, 284, 412,  15, 143, 308,  79, 244,
    372, 207,  47, 175, 340, 111, 276, 404,  31, 159,
    324,  95, 260, 388, 228,  63, 191, 356, 127, 292,
    420, 
    // Size 431
      0, 128, 303,  64, 239, 367, 192,  32, 160, 335,
     96, 271, 399,  16, 144, 319,  80, 255, 383, 223,
     48, 176, 351, 112, 287, 415, 208,   8, 136, 311,
     72, 247, 375, 200,  40, 168, 343, 104, 279, 407,
     24, 152, 327,  88, 263, 391, 231,  56, 184, 359,
    120, 295, 423, 212,   4, 132, 307,  68, 243, 371,
    196,  36, 164, 339, 100, 275, 403,  20, 148, 323,
     84, 259, 387, 227,  52, 180, 355, 116, 291, 419,
    219,  12, 140, 315,  76, 251, 379, 204,  44, 172,
    347, 108, 283, 411,  28, 156, 331,  92, 267, 395,
    235,  60, 188, 363, 124, 299, 427, 214,   2, 130,
    305,  66, 241, 369, 194,  34, 162, 337,  98, 273,
    401,  18, 146, 321,  82, 257, 385, 225,  50, 178,
    353, 114, 289, 417, 210,  10, 138, 313,  74, 249,
    377, 202,  42, 170, 345, 106, 281, 409,  26, 154,
    329,  90, 265, 393, 233,  58, 186, 361, 122, 297,
    425, 217,   6, 134, 309,  70, 245, 373, 198,  38,
    166, 341, 102, 277, 405,  22, 150, 325,  86, 261,
    389, 229,  54, 182, 357, 118, 293, 421, 221,  14,
    142, 317,  78, 253, 381, 206,  46, 174, 349, 110,
    285, 413,  30, 158, 333,  94, 269, 397, 237,  62,
    190, 365, 126, 301, 429, 215,   1, 129, 304,  65,
    240, 368, 193,  33, 161, 336,  97, 272, 400,  17,
    145, 320,  81, 256, 384, 224,  49, 177, 352, 113,
    288, 416, 209,   9, 137, 312,  73, 248, 376, 201,
     41, 169, 344, 105, 280, 408,  25, 153, 328,  89,
    264, 392, 232,  57, 185, 360, 121, 296, 424, 213,
      5, 133, 308,  69, 244, 372, 197,  37, 165, 340,
    101, 276, 404,  21, 149, 324,  85, 260, 388, 228,
     53, 181, 356, 117, 292, 420, 220,  13, 141, 316,
     77, 252, 380, 205,  45, 173, 348, 109, 284, 412,
     29, 157, 332,  93, 268, 396, 236,  61, 189, 364,
    125, 300, 428, 216,   3, 131, 306,  67, 242, 370,
    195,  35, 163, 338,  99, 274, 402,  19, 147, 322,
     83, 258, 386, 226,  51, 179, 354, 115, 290, 418,
    211,  11, 139, 314,  75, 250, 378, 203,  43, 171,
    346, 107, 282, 410,  27, 155, 330,  91, 266, 394,
    234,  59, 187, 362, 123, 298, 426, 218,   7, 135,
    310,  71, 246, 374, 199,  39, 167, 342, 103, 278,
    406,  23, 151, 326,  87, 262, 390, 230,  55, 183,
    358, 119, 294, 422, 222,  15, 143, 318,  79, 254,
    382, 207,  47, 175, 350, 111, 286, 414,  31, 159,
    334,  95, 270, 398, 238,  63, 191, 366, 127, 302,
    430, 
    // Size 433
      0, 128, 305,  64, 241, 369, 192,  32, 160, 337,
     96, 273, 401, 208,  16, 144, 321,  80, 257, 385,
    225,  48, 176, 353, 112, 289, 417,   8, 136, 313,
     72, 249, 377, 200,  40, 168, 345, 104, 281, 409,
    217,  24, 152, 329,  88, 265, 393, 233,  56, 184,
    361, 120, 297, 425,   4, 132, 309,  68, 245, 373,
    196,  36, 164, 341, 100, 277, 405, 212,  20, 148,
    325,  84, 261, 389, 229,  52, 180, 357, 116, 293,
    421,  12, 140, 317,  76, 253, 381, 204,  44, 172,
    349, 108, 285, 413, 221,  28, 156, 333,  92, 269,
    397, 237,  60, 188, 365, 124, 301, 429,   2, 130,
    307,  66, 243, 371, 194,  34, 162, 339,  98, 275,
    403, 210,  18, 146, 323,  82, 259, 387, 227,  50,
    178, 355, 114, 291, 419,  10, 138, 315,  74, 251,
    379, 202,  42, 170, 347, 106, 283, 411, 219,  26,
    154, 331,  90, 267, 395, 235,  58, 186, 363, 122,
    299, 427,   6, 134, 311,  70, 247, 375, 198,  38,
    166, 343, 102, 279, 407, 214,  22, 150, 327,  86,
    263, 391, 231,  54, 182, 359, 118, 295, 423,  14,
    142, 319,  78, 255, 383, 206,  46, 174, 351, 110,
    287, 415, 223,  30, 158, 335,  94, 271, 399, 239,
     62, 190, 367, 126, 303, 431, 216,   1, 129, 306,
     65, 242, 370, 193,  33, 161, 338,  97, 274, 402,
    209,  17, 145, 322,  81, 258, 386, 226,  49, 177,
    354, 113, 290, 418,   9, 137, 314,  73, 250, 378,
    201,  41, 169, 346, 105, 282, 410, 218,  25, 153,
    330,  89, 266, 394, 234,  57, 185, 362, 121, 298,
    426,   5, 133, 310,  69, 246, 374, 197,  37, 165,
    342, 101, 278, 406, 213,  21, 149, 326,  85, 262,
    390, 230,  53, 181, 358, 117, 294, 422,  13, 141,
    318,  77, 254, 382, 205,  45, 173, 350, 109, 286,
    414, 222,  29, 157, 334,  93, 270, 398, 238,  61,
    189, 366, 125, 302, 430,   3, 131, 308,  67, 244,
    372, 195,  35, 163, 340,  99, 276, 404, 211,  19,
    147, 324,  83, 260, 388, 228,  51, 179, 356, 115,
    292, 420,  11, 139, 316,  75, 252, 380, 203,  43,
    171, 348, 107, 284, 412, 220,  27, 155, 332,  91,
    268, 396, 236,  59, 187, 364, 123, 300, 428,   7,
    135, 312,  71, 248, 376, 199,  39, 167, 344, 103,
    280, 408, 215,  23, 151, 328,  87, 264, 392, 232,
     55, 183, 360, 119, 296, 424,  15, 143, 320,  79,
    256, 384, 207,  47, 175, 352, 111, 288, 416, 224,
     31, 159, 336,  95, 272, 400, 240,  63, 191, 368,
    127, 304, 432, 
    // Size 439
      0, 128, 311,  64, 247, 375, 192,  32, 160, 343,
     96, 279, 407, 208,  16, 144, 327,  80, 263, 391,
    231,  48, 176, 359, 112, 295, 423,   8, 136, 319,
     72, 255, 383, 200,  40, 168, 351, 104, 287, 415,
    223,  24, 152, 335,  88, 271, 399, 239,  56, 184,
    367, 120, 303, 431, 216,   4, 132, 315,  68, 251,
    379, 196,  36, 164, 347, 100, 283, 411, 212,  20,
    148, 331,  84, 267, 395, 235,  52, 180, 363, 116,
    299, 427,  12, 140, 323,  76, 259, 387, 204,  44,
    172, 355, 108, 291, 419, 227,  28, 156, 339,  92,
    275, 403, 243,  60, 188, 371, 124, 307, 435, 218,
      2, 130, 313,  66, 249, 377, 194,  34, 162, 345,
     98, 281, 409, 210,  18, 146, 329,  82, 265, 393,
    233,  50, 178, 361, 114, 297, 425,  10, 138, 321,
     74, 257, 385, 202,  42, 170, 353, 106, 289, 417,
    225,  26, 154, 337,  90, 273, 401, 241,  58, 186,
    369, 122, 305, 433, 221,   6, 134, 317,  70, 253,
    381, 198,  38, 166, 349, 102, 285, 413, 214,  22,
    150, 333,  86, 269, 397, 237,  54, 182, 365, 118,
    301, 429,  14, 142, 325,  78, 261, 389, 206,  46,
    174, 357, 110, 293, 421, 229,  30, 158, 341,  94,
    277, 405, 245,  62, 190, 373, 126, 309, 437, 219,
      1, 129, 312,  65, 248, 376, 193,  33, 161, 344,
     97, 280, 408, 209,  17, 145, 328,  81, 264, 392,
    232,  49, 177, 360, 113, 296, 424,   9, 137, 320,
     73, 256, 384, 201,  41, 169, 352, 105, 288, 416,
    224,  25, 153, 336,  89, 272, 400, 240,  57, 185,
    368, 121, 304, 432, 217,   5, 133, 316,  69, 252,
    380, 197,  37, 165, 348, 101, 284, 412, 213,  21,
    149, 332,  85, 268, 396, 236,  53, 181, 364, 117,
    300, 428,  13, 141, 324,  77, 260, 388, 205,  45,
    173, 356, 109, 292, 420, 228,  29, 157, 340,  93,
    276, 404, 244,  61, 189, 372, 125, 308, 436, 220,
      3, 131, 314,  67, 250, 378, 195,  35, 163, 346,
     99, 282, 410, 211,  19, 147, 330,  83, 266, 394,
    234,  51, 179, 362, 115, 298, 426,  11, 139, 322,
     75, 258, 386, 203,  43, 171, 354, 107, 290, 418,
    226,  27, 155, 338,  91, 274, 402, 242,  59, 187,
    370, 123, 306, 434, 222,   7, 135, 318,  71, 254,
    382, 199,  39, 167, 350, 103, 286, 414, 215,  23,
    151, 334,  87, 270, 398, 238,  55, 183, 366, 119,
    302, 430,  15, 143, 326,  79, 262, 390, 207,  47,
    175, 358, 111, 294, 422, 230,  31, 159, 342,  95,
    278, 406, 246,  63, 191, 374, 127, 310, 438, 
    // Size 443
      0, 128, 315,  64, 251, 379, 192,  32, 160, 347,
     96, 283, 411, 208,  16, 144, 331,  80, 267, 395,
    235,  48, 176, 363, 112, 299, 427, 216,   8, 136,
    323,  72, 259, 387, 200,  40, 168, 355, 104, 291,
    419, 227,  24, 152, 339,  88, 275, 403, 243,  56,
    184, 371, 120, 307, 435,   4, 132, 319,  68, 255,
    383, 196,  36, 164, 351, 100, 287, 415, 212,  20,
    148, 335,  84, 271, 399, 239,  52, 180, 367, 116,
    303, 431, 223,  12, 140, 327,  76, 263, 391, 204,
     44, 172, 359, 108, 295, 423, 231,  28, 156, 343,
     92, 279, 407, 247,  60, 188, 375, 124, 311, 439,
    220,   2, 130, 317,  66, 253, 381, 194,  34, 162,
    349,  98, 285, 413, 210,  18, 146, 333,  82, 269,
    397, 237,  50, 178, 365, 114, 301, 429, 218,  10,
    138, 325,  74, 261, 389, 202,  42, 170, 357, 106,
    293, 421, 229,  26, 154, 341,  90, 277, 405, 245,
     58, 186, 373, 122, 309, 437,   6, 134, 321,  70,
    257, 385, 198,  38, 166, 353, 102, 289, 417, 214,
     22, 150, 337,  86, 273, 401, 241,  54, 182, 369,
    118, 305, 433, 225,  14, 142, 329,  78, 265, 393,
    206,  46, 174, 361, 110, 297, 425, 233,  30, 158,
    345,  94, 281, 409, 249,  62, 190, 377, 126, 313,
    441, 221,   1, 129, 316,  65, 252, 380, 193,  33,
    161, 348,  97, 284, 412, 209,  17, 145, 332,  81,
    268, 396, 236,  49, 177, 364, 113, 300, 428, 217,
      9, 137, 324,  73, 260, 388, 201,  41, 169, 356,
    105, 292, 420, 228,  25, 153, 340,  89, 276, 404,
    244,  57, 185, 372, 121, 308, 436,   5, 133, 320,
     69, 256, 384, 197,  37, 165, 352, 101, 288, 416,
    213,  21, 149, 336,  85, 272, 400, 240,  53, 181,
    368, 117, 304, 432, 224,  13, 141, 328,  77, 264,
    392, 205,  45, 173, 360, 109, 296, 424, 232,  29,
    157, 344,  93, 280, 408, 248,  61, 189, 376, 125,
    312, 440, 222,   3, 131, 318,  67, 254, 382, 195,
     35, 163, 350,  99, 286, 414, 211,  19, 147, 334,
     83, 270, 398, 238,  51, 179, 366, 115, 302, 430,
    219,  11, 139, 326,  75, 262, 390, 203,  43, 171,
    358, 107, 294, 422, 230,  27, 155, 342,  91, 278,
    406, 246,  59, 187, 374, 123, 310, 438,   7, 135,
    322,  71, 258, 386, 199,  39, 167, 354, 103, 290,
    418, 215,  23, 151, 338,  87, 274, 402, 242,  55,
    183, 370, 119, 306, 434, 226,  15, 143, 330,  79,
    266, 394, 207,  47, 175, 362, 111, 298, 426, 234,
     31, 159, 346,  95, 282, 410, 250,  63, 191, 378,
    127, 314, 442, 
    // Size 449
      0, 128, 321, 192,  64, 257, 385,  32, 160, 353,
    225,  96, 289, 417,  16, 144, 337, 208,  80, 273,
    401,  48, 176, 369, 241, 112, 305, 433,   8, 136,
    329, 200,  72, 265, 393,  40, 168, 361, 233, 104,
    297, 425,  24, 152, 345, 216,  88, 281, 409,  56,
    184, 377, 249, 120, 313, 441,   4, 132, 325, 196,
     68, 261, 389,  36, 164, 357, 229, 100, 293, 421,
     20, 148, 341, 212,  84, 277, 405,  52, 180, 373,
    245, 116, 309, 437,  12, 140, 333, 204,  76, 269,
    397,  44, 172, 365, 237, 108, 301, 429,  28, 156,
    349, 220,  92, 285, 413,  60, 188, 381, 253, 124,
    317, 445,   2, 130, 323, 194,  66, 259, 387,  34,
    162, 355, 227,  98, 291, 419,  18, 146, 339, 210,
     82, 275, 403,  50, 178, 371, 243, 114, 307, 435,
     10, 138, 331, 202,  74, 267, 395,  42, 170, 363,
    235, 106, 299, 427,  26, 154, 347, 218,  90, 283,
    411,  58, 186, 379, 251, 122, 315, 443,   6, 134,
    327, 198,  70, 263, 391,  38, 166, 359, 231, 102,
    295, 423,  22, 150, 343, 214,  86, 279, 407,  54,
    182, 375, 247, 118, 311, 439,  14, 142, 335, 206,
     78, 271, 399,  46, 174, 367, 239, 110, 303, 431,
     30, 158, 351, 222,  94, 287, 415,  62, 190, 383,
    255, 126, 319, 447, 224,   1, 129, 322, 193,  65,
    258, 386,  33, 161, 354, 226,  97, 290, 418,  17,
    145, 338, 209,  81, 274, 402,  49, 177, 370, 242,
    113, 306, 434,   9, 137, 330, 201,  73, 266, 394,
     41, 169, 362, 234, 105, 298, 426,  25, 153, 346,
    217,  89, 282, 410,  57, 185, 378, 250, 121, 314,
    442,   5, 133, 326, 197,  69, 262, 390,  37, 165,
    358, 230, 101, 294, 422,  21, 149, 342, 213,  85,
    278, 406,  53, 181, 374, 246, 117, 310, 438,  13,
    141, 334, 205,  77, 270, 398,  45, 173, 366, 238,
    109, 302, 430,  29, 157, 350, 221,  93, 286, 414,
     61, 189, 382, 254, 125, 318, 446,   3, 131, 324,
    195,  67, 260, 388,  35, 163, 356, 228,  99, 292,
    420,  19, 147, 340, 211,  83, 276, 404,  51, 179,
    372, 244, 115, 308, 436,  11, 139, 332, 203,  75,
    268, 396,  43, 171, 364, 236, 107, 300, 428,  27,
    155, 348, 219,  91, 284, 412,  59, 187, 380, 252,
    123, 316, 444,   7, 135, 328, 199,  71, 264, 392,
     39, 167, 360, 232, 103, 296, 424,  23, 151, 344,
    215,  87, 280, 408,  55, 183, 376, 248, 119, 312,
    440,  15, 143, 336, 207,  79, 272, 400,  47, 175,
    368, 240, 111, 304, 432,  31, 159, 352, 223,  95,
    288, 416,  63, 191, 384, 256, 127, 320, 448, 
    // Size 457
      0, 128, 329, 192,  64, 265, 393,  32, 160, 361,
    233,  96, 297, 425,  16, 144, 345, 208,  80, 281,
    409,  48, 176, 377, 249, 112, 313, 441, 224,   8,
    136, 337, 200,  72, 273, 401,  40, 168, 369, 241,
    104, 305, 433,  24, 152, 353, 216,  88, 289, 417,
     56, 184, 385, 257, 120, 321, 449,   4, 132, 333,
    196,  68, 269, 397,  36, 164, 365, 237, 100, 301,
    429,  20, 148, 349, 212,  84, 285, 413,  52, 180,
    381, 253, 116, 317, 445, 229,  12, 140, 341, 204,
     76, 277, 405,  44, 172, 373, 245, 108, 309, 437,
     28, 156, 357, 220,  92, 293, 421,  60, 188, 389,
    261, 124, 325, 453,   2, 130, 331, 194,  66, 267,
    395,  34, 162, 363, 235,  98, 299, 427,  18, 146,
    347, 210,  82, 283, 411,  50, 178, 379, 251, 114,
    315, 443, 226,  10, 138, 339, 202,  74, 275, 403,
     42, 170, 371, 243, 106, 307, 435,  26, 154, 355,
    218,  90, 291, 419,  58, 186, 387, 259, 122, 323,
    451,   6, 134, 335, 198,  70, 271, 399,  38, 166,
    367, 239, 102, 303, 431,  22, 150, 351, 214,  86,
    287, 415,  54, 182, 383, 255, 118, 319, 447, 231,
     14, 142, 343, 206,  78, 279, 407,  46, 174, 375,
    247, 110, 311, 439,  30, 158, 359, 222,  94, 295,
    423,  62, 190, 391, 263, 126, 327, 455, 228,   1,
    129, 330, 193,  65, 266, 394,  33, 161, 362, 234,
     97, 298, 426,  17, 145, 346, 209,  81, 282, 410,
     49, 177, 378, 250, 113, 314, 442, 225,   9, 137,
    338, 201,  73, 274, 402,  41, 169, 370, 242, 105,
    306, 434,  25, 153, 354, 217,  89, 290, 418,  57,
    185, 386, 258, 121, 322, 450,   5, 133, 334, 197,
     69, 270, 398,  37, 165, 366, 238, 101, 302, 430,
     21, 149, 350, 213,  85, 286, 414,  53, 181, 382,
    254, 117, 318, 446, 230,  13, 141, 342, 205,  77,
    278, 406,  45, 173, 374, 246, 109, 310, 438,  29,
    157, 358, 221,  93, 294, 422,  61, 189, 390, 262,
    125, 326, 454,   3, 131, 332, 195,  67, 268, 396,
     35, 163, 364, 236,  99, 300, 428,  19, 147, 348,
    211,  83, 284, 412,  51, 179, 380, 252, 115, 316,
    444, 227,  11, 139, 340, 203,  75, 276, 404,  43,
    171, 372, 244, 107, 308, 436,  27, 155, 356, 219,
     91, 292, 420,  59, 187, 388, 260, 123, 324, 452,
      7, 135, 336, 199,  71, 272, 400,  39, 167, 368,
    240, 103, 304, 432,  23, 151, 352, 215,  87, 288,
    416,  55, 183, 384, 256, 119, 320, 448, 232,  15,
    143, 344, 207,  79, 280, 408,  47, 175, 376, 248,
    111, 312, 440,  31, 159, 360, 223,  95, 296, 424,
     63, 191, 392, 264, 127, 328, 456, 
    // Size 461
      0, 128, 333, 192,  64, 269, 397,  32, 160, 365,
    237,  96, 301, 429,  16, 144, 349, 208,  80, 285,
    413,  48, 176, 381, 253, 112, 317, 445, 224,   8,
    136, 341, 200,  72, 277, 405,  40, 168, 373, 245,
    104, 309, 437,  24, 152, 357, 216,  88, 293, 421,
     56, 184, 389, 261, 120, 325, 453, 228,   4, 132,
    337, 196,  68, 273, 401,  36, 164, 369, 241, 100,
    305, 433,  20, 148, 353, 212,  84, 289, 417,  52,
    180, 385, 257, 116, 321, 449, 233,  12, 140, 345,
    204,  76, 281, 409,  44, 172, 377, 249, 108, 313,
    441,  28, 156, 361, 220,  92, 297, 425,  60, 188,
    393, 265, 124, 329, 457,   2, 130, 335, 194,  66,
    271, 399,  34, 162, 367, 239,  98, 303, 431,  18,
    146, 351, 210,  82, 287, 415,  50, 178, 383, 255,
    114, 319, 447, 226,  10, 138, 343, 202,  74, 279,
    407,  42, 170, 375, 247, 106, 311, 439,  26, 154,
    359, 218,  90, 295, 423,  58, 186, 391, 263, 122,
    327, 455, 231,   6, 134, 339, 198,  70, 275, 403,
     38, 166, 371, 243, 102, 307, 435,  22, 150, 355,
    214,  86, 291, 419,  54, 182, 387, 259, 118, 323,
    451, 235,  14, 142, 347, 206,  78, 283, 411,  46,
    174, 379, 251, 110, 315, 443,  30, 158, 363, 222,
     94, 299, 427,  62, 190, 395, 267, 126, 331, 459,
    230,   1, 129, 334, 193,  65, 270, 398,  33, 161,
    366, 238,  97, 302, 430,  17, 145, 350, 209,  81,
    286, 414,  49, 177, 382, 254, 113, 318, 446, 225,
      9, 137, 342, 201,  73, 278, 406,  41, 169, 374,
    246, 105, 310, 438,  25, 153, 358, 217,  89, 294,
    422,  57, 185, 390, 262, 121, 326, 454, 229,   5,
    133, 338, 197,  69, 274, 402,  37, 165, 370, 242,
    101, 306, 434,  21, 149, 354, 213,  85, 290, 418,
     53, 181, 386, 258, 117, 322, 450, 234,  13, 141,
    346, 205,  77, 282, 410,  45, 173, 378, 250, 109,
    314, 442,  29, 157, 362, 221,  93, 298, 426,  61,
    189, 394, 266, 125, 330, 458,   3, 131, 336, 195,
     67, 272, 400,  35, 163, 368, 240,  99, 304, 432,
     19, 147, 352, 211,  83, 288, 416,  51, 179, 384,
    256, 115, 320, 448, 227,  11, 139, 344, 203,  75,
    280, 408,  43, 171, 376, 248, 107, 312, 440,  27,
    155, 360, 219,  91, 296, 424,  59, 187, 392, 264,
    123, 328, 456, 232,   7, 135, 340, 199,  71, 276,
    404,  39, 167, 372, 244, 103, 308, 436,  23, 151,
    356, 215,  87, 292, 420,  55, 183, 388, 260, 119,
    324, 452, 236,  15, 143, 348, 207,  79, 284, 412,
     47, 175, 380, 252, 111, 316, 444,  31, 159, 364,
    223,  95, 300, 428,  63, 191, 396, 268, 127, 332,
    460, 
    // Size 463
      0, 128, 335, 192,  64, 271, 399,  32, 160, 367,
    239,  96, 303, 431,  16, 144, 351, 208,  80, 287,
    415,  48, 176, 383, 255, 112, 319, 447, 224,   8,
    136, 343, 200,  72, 279, 407,  40, 168, 375, 247,
    104, 311, 439,  24, 152, 359, 216,  88, 295, 423,
     56, 184, 391, 263, 120, 327, 455, 228,   4, 132,
    339, 196,  68, 275, 403,  36, 164, 371, 243, 100,
    307, 435,  20, 148, 355, 212,  84, 291, 419,  52,
    180, 387, 259, 116, 323, 451, 235,  12, 140, 347,
    204,  76, 283, 411,  44, 172, 379, 251, 108, 315,
    443,  28, 156, 363, 220,  92, 299, 427,  60, 188,
    395, 267, 124, 331, 459, 230,   2, 130, 337, 194,
     66, 273, 401,  34, 162, 369, 241,  98, 305, 433,
     18, 146, 353, 210,  82, 289, 417,  50, 178, 385,
    257, 114, 321, 449, 226,  10, 138, 345, 202,  74,
    281, 409,  42, 170, 377, 249, 106, 313, 441,  26,
    154, 361, 218,  90, 297, 425,  58, 186, 393, 265,
    122, 329, 457, 233,   6, 134, 341, 198,  70, 277,
    405,  38, 166, 373, 245, 102, 309, 437,  22, 150,
    357, 214,  86, 293, 421,  54, 182, 389, 261, 118,
    325, 453, 237,  14, 142, 349, 206,  78, 285, 413,
     46, 174, 381, 253, 110, 317, 445,  30, 158, 365,
    222,  94, 301, 429,  62, 190, 397, 269, 126, 333,
    461, 231,   1, 129, 336, 193,  65, 272, 400,  33,
    161, 368, 240,  97, 304, 432,  17, 145, 352, 209,
     81, 288, 416,  49, 177, 384, 256, 113, 320, 448,
    225,   9, 137, 344, 201,  73, 280, 408,  41, 169,
    376, 248, 105, 312, 440,  25, 153, 360, 217,  89,
    296, 424,  57, 185, 392, 264, 121, 328, 456, 229,
      5, 133, 340, 197,  69, 276, 404,  37, 165, 372,
    244, 101, 308, 436,  21, 149, 356, 213,  85, 292,
    420,  53, 181, 388, 260, 117, 324, 452, 236,  13,
    141, 348, 205,  77, 284, 412,  45, 173, 380, 252,
    109, 316, 444,  29, 157, 364, 221,  93, 300, 428,
     61, 189, 396, 268, 125, 332, 460, 232,   3, 131,
    338, 195,  67, 274, 402,  35, 163, 370, 242,  99,
    306, 434,  19, 147, 354, 211,  83, 290, 418,  51,
    179, 386, 258, 115, 322, 450, 227,  11, 139, 346,
    203,  75, 282, 410,  43, 171, 378, 250, 107, 314,
    442,  27, 155, 362, 219,  91, 298, 426,  59, 187,
    394, 266, 123, 330, 458, 234,   7, 135, 342, 199,
     71, 278, 406,  39, 167, 374, 246, 103, 310, 438,
     23, 151, 358, 215,  87, 294, 422,  55, 183, 390,
    262, 119, 326, 454, 238,  15, 143, 350, 207,  79,
    286, 414,  47, 175, 382, 254, 111, 318, 446,  31,
    159, 366, 223,  95, 302, 430,  63, 191, 398, 270,
    127, 334, 462, 
    // Size 467
      0, 128, 339, 192,  64, 275, 403,  32, 160, 371,
    243,  96, 307, 435, 224,  16, 144, 355, 208,  80,
    291, 419,  48, 176, 387, 259, 112, 323, 451,   8,
    136, 347, 200,  72, 283, 411,  40, 168, 379, 251,
    104, 315, 443, 235,  24, 152, 363, 216,  88, 299,
    427,  56, 184, 395, 267, 120, 331, 459,   4, 132,
    343, 196,  68, 279, 407,  36, 164, 375, 247, 100,
    311, 439, 228,  20, 148, 359, 212,  84, 295, 423,
     52, 180, 391, 263, 116, 327, 455,  12, 140, 351,
    204,  76, 287, 415,  44, 172, 383, 255, 108, 319,
    447, 239,  28, 156, 367, 220,  92, 303, 431,  60,
    188, 399, 271, 124, 335, 463, 232,   2, 130, 341,
    194,  66, 277, 405,  34, 162, 373, 245,  98, 309,
    437, 226,  18, 146, 357, 210,  82, 293, 421,  50,
    178, 389, 261, 114, 325, 453,  10, 138, 349, 202,
     74, 285, 413,  42, 170, 381, 253, 106, 317, 445,
    237,  26, 154, 365, 218,  90, 301, 429,  58, 186,
    397, 269, 122, 333, 461,   6, 134, 345, 198,  70,
    281, 409,  38, 166, 377, 249, 102, 313, 441, 230,
     22, 150, 361, 214,  86, 297, 425,  54, 182, 393,
    265, 118, 329, 457,  14, 142, 353, 206,  78, 289,
    417,  46, 174, 385, 257, 110, 321, 449, 241,  30,
    158, 369, 222,  94, 305, 433,  62, 190, 401, 273,
    126, 337, 465, 233,   1, 129, 340, 193,  65, 276,
    404,  33, 161, 372, 244,  97, 308, 436, 225,  17,
    145, 356, 209,  81, 292, 420,  49, 177, 388, 260,
    113, 324, 452,   9, 137, 348, 201,  73, 284, 412,
     41, 169, 380, 252, 105, 316, 444, 236,  25, 153,
    364, 217,  89, 300, 428,  57, 185, 396, 268, 121,
    332, 460,   5, 133, 344, 197,  69, 280, 408,  37,
    165, 376, 248, 101, 312, 440, 229,  21, 149, 360,
    213,  85, 296, 424,  53, 181, 392, 264, 117, 328,
    456,  13, 141, 352, 205,  77, 288, 416,  45, 173,
    384, 256, 109, 320, 448, 240,  29, 157, 368, 221,
     93, 304, 432,  61, 189, 400, 272, 125, 336, 464,
    234,   3, 131, 342, 195,  67, 278, 406,  35, 163,
    374, 246,  99, 310, 438, 227,  19, 147, 358, 211,
     83, 294, 422,  51, 179, 390, 262, 115, 326, 454,
     11, 139, 350, 203,  75, 286, 414,  43, 171, 382,
    254, 107, 318, 446, 238,  27, 155, 366, 219,  91,
    302, 430,  59, 187, 398, 270, 123, 334, 462,   7,
    135, 346, 199,  71, 282, 410,  39, 167, 378, 250,
    103, 314, 442, 231,  23, 151, 362, 215,  87, 298,
    426,  55, 183, 394, 266, 119, 330, 458,  15, 143,
    354, 207,  79, 290, 418,  47, 175, 386, 258, 111,
    322, 450, 242,  31, 159, 370, 223,  95, 306, 434,
     63, 191, 402, 274, 127, 338, 466, 
    // Size 479
      0, 128, 351, 192,  64, 287, 415,  32, 160, 383,
    255,  96, 319, 447, 224,  16, 144, 367, 208,  80,
    303, 431,  48, 176, 399, 271, 112, 335, 463, 232,
      8, 136, 359, 200,  72, 295, 423,  40, 168, 391,
    263, 104, 327, 455, 247,  24, 152, 375, 216,  88,
    311, 439,  56, 184, 407, 279, 120, 343, 471, 236,
      4, 132, 355, 196,  68, 291, 419,  36, 164, 387,
    259, 100, 323, 451, 228,  20, 148, 371, 212,  84,
    307, 435,  52, 180, 403, 275, 116, 339, 467, 243,
     12, 140, 363, 204,  76, 299, 427,  44, 172, 395,
    267, 108, 331, 459, 251,  28, 156, 379, 220,  92,
    315, 443,  60, 188, 411, 283, 124, 347, 475, 238,
      2, 130, 353, 194,  66, 289, 417,  34, 162, 385,
    257,  98, 321, 449, 226,  18, 146, 369, 210,  82,
    305, 433,  50, 178, 401, 273, 114, 337, 465, 234,
     10, 138, 361, 202,  74, 297, 425,  42, 170, 393,
    265, 106, 329, 457, 249,  26, 154, 377, 218,  90,
    313, 441,  58, 186, 409, 281, 122, 345, 473, 241,
      6, 134, 357, 198,  70, 293, 421,  38, 166, 389,
    261, 102, 325, 453, 230,  22, 150, 373, 214,  86,
    309, 437,  54, 182, 405, 277, 118, 341, 469, 245,
     14, 142, 365, 206,  78, 301, 429,  46, 174, 397,
    269, 110, 333, 461, 253,  30, 158, 381, 222,  94,
    317, 445,  62, 190, 413, 285, 126, 349, 477, 239,
      1, 129, 352, 193,  65, 288, 416,  33, 161, 384,
    256,  97, 320, 448, 225,  17, 145, 368, 209,  81,
    304, 432,  49, 177, 400, 272, 113, 336, 464, 233,
      9, 137, 360, 201,  73, 296, 424,  41, 169, 392,
    264, 105, 328, 456, 248,  25, 153, 376, 217,  89,
    312, 440,  57, 185, 408, 280, 121, 344, 472, 237,
      5, 133, 356, 197,  69, 292, 420,  37, 165, 388,
    260, 101, 324, 452, 229,  21, 149, 372, 213,  85,
    308, 436,  53, 181, 404, 276, 117, 340, 468, 244,
     13, 141, 364, 205,  77, 300, 428,  45, 173, 396,
    268, 109, 332, 460, 252,  29, 157, 380, 221,  93,
    316, 444,  61, 189, 412, 284, 125, 348, 476, 240,
      3, 131, 354, 195,  67, 290, 418,  35, 163, 386,
    258,  99, 322, 450, 227,  19, 147, 370, 211,  83,
    306, 434,  51, 179, 402, 274, 115, 338, 466, 235,
     11, 139, 362, 203,  75, 298, 426,  43, 171, 394,
    266, 107, 330, 458, 250,  27, 155, 378, 219,  91,
    314, 442,  59, 187, 410, 282, 123, 346, 474, 242,
      7, 135, 358, 199,  71, 294, 422,  39, 167, 390,
    262, 103, 326, 454, 231,  23, 151, 374, 215,  87,
    310, 438,  55, 183, 406, 278, 119, 342, 470, 246,
     15, 143, 366, 207,  79, 302, 430,  47, 175, 398,
    270, 111, 334, 462, 254,  31, 159, 382, 223,  95,
    318, 446,  63, 191, 414, 286, 127, 350, 478, 
    // Size 487
      0, 128, 359, 192,  64, 295, 423, 224,  32, 160,
    391, 263,  96, 327, 455,  16, 144, 375, 208,  80,
    311, 439, 247,  48, 176, 407, 279, 112, 343, 471,
      8, 136, 367, 200,  72, 303, 431, 232,  40, 168,
    399, 271, 104, 335, 463,  24, 152, 383, 216,  88,
    319, 447, 255,  56, 184, 415, 287, 120, 351, 479,
    240,   4, 132, 363, 196,  68, 299, 427, 228,  36,
    164, 395, 267, 100, 331, 459,  20, 148, 379, 212,
     84, 315, 443, 251,  52, 180, 411, 283, 116, 347,
    475,  12, 140, 371, 204,  76, 307, 435, 236,  44,
    172, 403, 275, 108, 339, 467,  28, 156, 387, 220,
     92, 323, 451, 259,  60, 188, 419, 291, 124, 355,
    483, 242,   2, 130, 361, 194,  66, 297, 425, 226,
     34, 162, 393, 265,  98, 329, 457,  18, 146, 377,
    210,  82, 313, 441, 249,  50, 178, 409, 281, 114,
    345, 473,  10, 138, 369, 202,  74, 305, 433, 234,
     42, 170, 401, 273, 106, 337, 465,  26, 154, 385,
    218,  90, 321, 449, 257,  58, 186, 417, 289, 122,
    353, 481, 245,   6, 134, 365, 198,  70, 301, 429,
    230,  38, 166, 397, 269, 102, 333, 461,  22, 150,
    381, 214,  86, 317, 445, 253,  54, 182, 413, 285,
    118, 349, 477,  14, 142, 373, 206,  78, 309, 437,
    238,  46, 174, 405, 277, 110, 341, 469,  30, 158,
    389, 222,  94, 325, 453, 261,  62, 190, 421, 293,
    126, 357, 485, 243,   1, 129, 360, 193,  65, 296,
    424, 225,  33, 161, 392, 264,  97, 328, 456,  17,
    145, 376, 209,  81, 312, 440, 248,  49, 177, 408,
    280, 113, 344, 472,   9, 137, 368, 201,  73, 304,
    432, 233,  41, 169, 400, 272, 105, 336, 464,  25,
    153, 384, 217,  89, 320, 448, 256,  57, 185, 416,
    288, 121, 352, 480, 241,   5, 133, 364, 197,  69,
    300, 428, 229,  37, 165, 396, 268, 101, 332, 460,
     21, 149, 380, 213,  85, 316, 444, 252,  53, 181,
    412, 284, 117, 348, 476,  13, 141, 372, 205,  77,
    308, 436, 237,  45, 173, 404, 276, 109, 340, 468,
     29, 157, 388, 221,  93, 324, 452, 260,  61, 189,
    420, 292, 125, 356, 484, 244,   3, 131, 362, 195,
     67, 298, 426, 227,  35, 163, 394, 266,  99, 330,
    458,  19, 147, 378, 211,  83, 314, 442, 250,  51,
    179, 410, 282, 115, 346, 474,  11, 139, 370, 203,
     75, 306, 434, 235,  43, 171, 402, 274, 107, 338,
    466,  27, 155, 386, 219,  91, 322, 450, 258,  59,
    187, 418, 290, 123, 354, 482, 246,   7, 135, 366,
    199,  71, 302, 430, 231,  39, 167, 398, 270, 103,
    334, 462,  23, 151, 382, 215,  87, 318, 446, 254,
     55, 183, 414, 286, 119, 350, 478,  15, 143, 374,
    207,  79, 310, 438, 239,  47, 175, 406, 278, 111,
    342, 470,  31, 159, 390, 223,  95, 326, 454, 262,
     63, 191, 422, 294, 127, 358, 486, 
    // Size 491
      0, 128, 363, 192,  64, 299, 427, 224,  32, 160,
    395, 267,  96, 331, 459,  16, 144, 379, 208,  80,
    315, 443, 251,  48, 176, 411, 283, 112, 347, 475,
    240,   8, 136, 371, 200,  72, 307, 435, 232,  40,
    168, 403, 275, 104, 339, 467,  24, 152, 387, 216,
     88, 323, 451, 259,  56, 184, 419, 291, 120, 355,
    483,   4, 132, 367, 196,  68, 303, 431, 228,  36,
    164, 399, 271, 100, 335, 463,  20, 148, 383, 212,
     84, 319, 447, 255,  52, 180, 415, 287, 116, 351,
    479, 247,  12, 140, 375, 204,  76, 311, 439, 236,
     44, 172, 407, 279, 108, 343, 471,  28, 156, 391,
    220,  92, 327, 455, 263,  60, 188, 423, 295, 124,
    359, 487, 244,   2, 130, 365, 194,  66, 301, 429,
    226,  34, 162, 397, 269,  98, 333, 461,  18, 146,
    381, 210,  82, 317, 445, 253,  50, 178, 413, 285,
    114, 349, 477, 242,  10, 138, 373, 202,  74, 309,
    437, 234,  42, 170, 405, 277, 106, 341, 469,  26,
    154, 389, 218,  90, 325, 453, 261,  58, 186, 421,
    293, 122, 357, 485,   6, 134, 369, 198,  70, 305,
    433, 230,  38, 166, 401, 273, 102, 337, 465,  22,
    150, 385, 214,  86, 321, 449, 257,  54, 182, 417,
    289, 118, 353, 481, 249,  14, 142, 377, 206,  78,
    313, 441, 238,  46, 174, 409, 281, 110, 345, 473,
     30, 158, 393, 222,  94, 329, 457, 265,  62, 190,
    425, 297, 126, 361, 489, 245,   1, 129, 364, 193,
     65, 300, 428, 225,  33, 161, 396, 268,  97, 332,
    460,  17, 145, 380, 209,  81, 316, 444, 252,  49,
    177, 412, 284, 113, 348, 476, 241,   9, 137, 372,
    201,  73, 308, 436, 233,  41, 169, 404, 276, 105,
    340, 468,  25, 153, 388, 217,  89, 324, 452, 260,
     57, 185, 420, 292, 121, 356, 484,   5, 133, 368,
    197,  69, 304, 432, 229,  37, 165, 400, 272, 101,
    336, 464,  21, 149, 384, 213,  85, 320, 448, 256,
     53, 181, 416, 288, 117, 352, 480, 248,  13, 141,
    376, 205,  77, 312, 440, 237,  45, 173, 408, 280,
    109, 344, 472,  29, 157, 392, 221,  93, 328, 456,
    264,  61, 189, 424, 296, 125, 360, 488, 246,   3,
    131, 366, 195,  67, 302, 430, 227,  35, 163, 398,
    270,  99, 334, 462,  19, 147, 382, 211,  83, 318,
    446, 254,  51, 179, 414, 286, 115, 350, 478, 243,
     11, 139, 374, 203,  75, 310, 438, 235,  43, 171,
    406, 278, 107, 342, 470,  27, 155, 390, 219,  91,
    326, 454, 262,  59, 187, 422, 294, 123, 358, 486,
      7, 135, 370, 199,  71, 306, 434, 231,  39, 167,
    402, 274, 103, 338, 466,  23, 151, 386, 215,  87,
    322, 450, 258,  55, 183, 418, 290, 119, 354, 482,
    250,  15, 143, 378, 207,  79, 314, 442, 239,  47,
    175, 410, 282, 111, 346, 474,  31, 159, 394, 223,
     95, 330, 458, 266,  63, 191, 426, 298, 127, 362,
    490, 
    // Size 499
      0, 128, 371, 192,  64, 307, 435, 224,  32, 160,
    403, 275,  96, 339, 467, 240,  16, 144, 387, 208,
     80, 323, 451, 259,  48, 176, 419, 291, 112, 355,
    483,   8, 136, 379, 200,  72, 315, 443, 232,  40,
    168, 411, 283, 104, 347, 475, 251,  24, 152, 395,
    216,  88, 331, 459, 267,  56, 184, 427, 299, 120,
    363, 491,   4, 132, 375, 196,  68, 311, 439, 228,
     36, 164, 407, 279, 100, 343, 471, 244,  20, 148,
    391, 212,  84, 327, 455, 263,  52, 180, 423, 295,
    116, 359, 487,  12, 140, 383, 204,  76, 319, 447,
    236,  44, 172, 415, 287, 108, 351, 479, 255,  28,
    156, 399, 220,  92, 335, 463, 271,  60, 188, 431,
    303, 124, 367, 495, 248,   2, 130, 373, 194,  66,
    309, 437, 226,  34, 162, 405, 277,  98, 341, 469,
    242,  18, 146, 389, 210,  82, 325, 453, 261,  50,
    178, 421, 293, 114, 357, 485,  10, 138, 381, 202,
     74, 317, 445, 234,  42, 170, 413, 285, 106, 349,
    477, 253,  26, 154, 397, 218,  90, 333, 461, 269,
     58, 186, 429, 301, 122, 365, 493,   6, 134, 377,
    198,  70, 313, 441, 230,  38, 166, 409, 281, 102,
    345, 473, 246,  22, 150, 393, 214,  86, 329, 457,
    265,  54, 182, 425, 297, 118, 361, 489,  14, 142,
    385, 206,  78, 321, 449, 238,  46, 174, 417, 289,
    110, 353, 481, 257,  30, 158, 401, 222,  94, 337,
    465, 273,  62, 190, 433, 305, 126, 369, 497, 249,
      1, 129, 372, 193,  65, 308, 436, 225,  33, 161,
    404, 276,  97, 340, 468, 241,  17, 145, 388, 209,
     81, 324, 452, 260,  49, 177, 420, 292, 113, 356,
    484,   9, 137, 380, 201,  73, 316, 444, 233,  41,
    169, 412, 284, 105, 348, 476, 252,  25, 153, 396,
    217,  89, 332, 460, 268,  57, 185, 428, 300, 121,
    364, 492,   5, 133, 376, 197,  69, 312, 440, 229,
     37, 165, 408, 280, 101, 344, 472, 245,  21, 149,
    392, 213,  85, 328, 456, 264,  53, 181, 424, 296,
    117, 360, 488,  13, 141, 384, 205,  77, 320, 448,
    237,  45, 173, 416, 288, 109, 352, 480, 256,  29,
    157, 400, 221,  93, 336, 464, 272,  61, 189, 432,
    304, 125, 368, 496, 250,   3, 131, 374, 195,  67,
    310, 438, 227,  35, 163, 406, 278,  99, 342, 470,
    243,  19, 147, 390, 211,  83, 326, 454, 262,  51,
    179, 422, 294, 115, 358, 486,  11, 139, 382, 203,
     75, 318, 446, 235,  43, 171, 414, 286, 107, 350,
    478, 254,  27, 155, 398, 219,  91, 334, 462, 270,
     59, 187, 430, 302, 123, 366, 494,   7, 135, 378,
    199,  71, 314, 442, 231,  39, 167, 410, 282, 103,
    346, 474, 247,  23, 151, 394, 215,  87, 330, 458,
    266,  55, 183, 426, 298, 119, 362, 490,  15, 143,
    386, 207,  79, 322, 450, 239,  47, 175, 418, 290,
    111, 354, 482, 258,  31, 159, 402, 223,  95, 338,
    466, 274,  63, 191, 434, 306, 127, 370, 498, 
    // Size 503
      0, 128, 375, 192,  64, 311, 439, 224,  32, 160,
    407, 279,  96, 343, 471, 240,  16, 144, 391, 208,
     80, 327, 455, 263,  48, 176, 423, 295, 112, 359,
    487,   8, 136, 383, 200,  72, 319, 447, 232,  40,
    168, 415, 287, 104, 351, 479, 255,  24, 152, 399,
    216,  88, 335, 463, 271,  56, 184, 431, 303, 120,
    367, 495, 248,   4, 132, 379, 196,  68, 315, 443,
    228,  36, 164, 411, 283, 100, 347, 475, 244,  20,
    148, 395, 212,  84, 331, 459, 267,  52, 180, 427,
    299, 116, 363, 491,  12, 140, 387, 204,  76, 323,
    451, 236,  44, 172, 419, 291, 108, 355, 483, 259,
     28, 156, 403, 220,  92, 339, 467, 275,  60, 188,
    435, 307, 124, 371, 499, 250,   2, 130, 377, 194,
     66, 313, 441, 226,  34, 162, 409, 281,  98, 345,
    473, 242,  18, 146, 393, 210,  82, 329, 457, 265,
     50, 178, 425, 297, 114, 361, 489,  10, 138, 385,
    202,  74, 321, 449, 234,  42, 170, 417, 289, 106,
    353, 481, 257,  26, 154, 401, 218,  90, 337, 465,
    273,  58, 186, 433, 305, 122, 369, 497, 253,   6,
    134, 381, 198,  70, 317, 445, 230,  38, 166, 413,
    285, 102, 349, 477, 246,  22, 150, 397, 214,  86,
    333, 461, 269,  54, 182, 429, 301, 118, 365, 493,
     14, 142, 389, 206,  78, 325, 453, 238,  46, 174,
    421, 293, 110, 357, 485, 261,  30, 158, 405, 222,
     94, 341, 469, 277,  62, 190, 437, 309, 126, 373,
    501, 251,   1, 129, 376, 193,  65, 312, 440, 225,
     33, 161, 408, 280,  97, 344, 472, 241,  17, 145,
    392, 209,  81, 328, 456, 264,  49, 177, 424, 296,
    113, 360, 488,   9, 137, 384, 201,  73, 320, 448,
    233,  41, 169, 416, 288, 105, 352, 480, 256,  25,
    153, 400, 217,  89, 336, 464, 272,  57, 185, 432,
    304, 121, 368, 496, 249,   5, 133, 380, 197,  69,
    316, 444, 229,  37, 165, 412, 284, 101, 348, 476,
    245,  21, 149, 396, 213,  85, 332, 460, 268,  53,
    181, 428, 300, 117, 364, 492,  13, 141, 388, 205,
     77, 324, 452, 237,  45, 173, 420, 292, 109, 356,
    484, 260,  29, 157, 404, 221,  93, 340, 468, 276,
     61, 189, 436, 308, 125, 372, 500, 252,   3, 131,
    378, 195,  67, 314, 442, 227,  35, 163, 410, 282,
     99, 346, 474, 243,  19, 147, 394, 211,  83, 330,
    458, 266,  51, 179, 426, 298, 115, 362, 490,  11,
    139, 386, 203,  75, 322, 450, 235,  43, 171, 418,
    290, 107, 354, 482, 258,  27, 155, 402, 219,  91,
    338, 466, 274,  59, 187, 434, 306, 123, 370, 498,
    254,   7, 135, 382, 199,  71, 318, 446, 231,  39,
    167, 414, 286, 103, 350, 478, 247,  23, 151, 398,
    215,  87, 334, 462, 270,  55, 183, 430, 302, 119,
    366, 494,  15, 143, 390, 207,  79, 326, 454, 239,
     47, 175, 422, 294, 111, 358, 486, 262,  31, 159,
    406, 223,  95, 342, 470, 278,  63, 191, 438, 310,
    127, 374, 502, 
    // Size 509
      0, 128, 381, 192,  64, 317, 445, 224,  32, 160,
    413, 285,  96, 349, 477, 240,  16, 144, 397, 208,
     80, 333, 461, 269,  48, 176, 429, 301, 112, 365,
    493, 248,   8, 136, 389, 200,  72, 325, 453, 232,
     40, 168, 421, 293, 104, 357, 485, 261,  24, 152,
    405, 216,  88, 341, 469, 277,  56, 184, 437, 309,
    120, 373, 501, 252,   4, 132, 385, 196,  68, 321,
    449, 228,  36, 164, 417, 289, 100, 353, 481, 244,
     20, 148, 401, 212,  84, 337, 465, 273,  52, 180,
    433, 305, 116, 369, 497, 257,  12, 140, 393, 204,
     76, 329, 457, 236,  44, 172, 425, 297, 108, 361,
    489, 265,  28, 156, 409, 220,  92, 345, 473, 281,
     60, 188, 441, 313, 124, 377, 505,   2, 130, 383,
    194,  66, 319, 447, 226,  34, 162, 415, 287,  98,
    351, 479, 242,  18, 146, 399, 210,  82, 335, 463,
    271,  50, 178, 431, 303, 114, 367, 495, 250,  10,
    138, 391, 202,  74, 327, 455, 234,  42, 170, 423,
    295, 106, 359, 487, 263,  26, 154, 407, 218,  90,
    343, 471, 279,  58, 186, 439, 311, 122, 375, 503,
    255,   6, 134, 387, 198,  70, 323, 451, 230,  38,
    166, 419, 291, 102, 355, 483, 246,  22, 150, 403,
    214,  86, 339, 467, 275,  54, 182, 435, 307, 118,
    371, 499, 259,  14, 142, 395, 206,  78, 331, 459,
    238,  46, 174, 427, 299, 110, 363, 491, 267,  30,
    158, 411, 222,  94, 347, 475, 283,  62, 190, 443,
    315, 126, 379, 507, 254,   1, 129, 382, 193,  65,
    318, 446, 225,  33, 161, 414, 286,  97, 350, 478,
    241,  17, 145, 398, 209,  81, 334, 462, 270,  49,
    177, 430, 302, 113, 366, 494, 249,   9, 137, 390,
    201,  73, 326, 454, 233,  41, 169, 422, 294, 105,
    358, 486, 262,  25, 153, 406, 217,  89, 342, 470,
    278,  57, 185, 438, 310, 121, 374, 502, 253,   5,
    133, 386, 197,  69, 322, 450, 229,  37, 165, 418,
    290, 101, 354, 482, 245,  21, 149, 402, 213,  85,
    338, 466, 274,  53, 181, 434, 306, 117, 370, 498,
    258,  13, 141, 394, 205,  77, 330, 458, 237,  45,
    173, 426, 298, 109, 362, 490, 266,  29, 157, 410,
    221,  93, 346, 474, 282,  61, 189, 442, 314, 125,
    378, 506,   3, 131, 384, 195,  67, 320, 448, 227,
     35, 163, 416, 288,  99, 352, 480, 243,  19, 147,
    400, 211,  83, 336, 464, 272,  51, 179, 432, 304,
    115, 368, 496, 251,  11, 139, 392, 203,  75, 328,
    456, 235,  43, 171, 424, 296, 107, 360, 488, 264,
     27, 155, 408, 219,  91, 344, 472, 280,  59, 187,
    440, 312, 123, 376, 504, 256,   7, 135, 388, 199,
     71, 324, 452, 231,  39, 167, 420, 292, 103, 356,
    484, 247,  23, 151, 404, 215,  87, 340, 468, 276,
     55, 183, 436, 308, 119, 372, 500, 260,  15, 143,
    396, 207,  79, 332, 460, 239,  47, 175, 428, 300,
    111, 364, 492, 268,  31, 159, 412, 223,  95, 348,
    476, 284,  63, 191, 444, 316, 127, 380, 508, 
    // Size 521
      0, 265, 128, 393,  64, 329, 192, 457,  32, 297,
    160, 425,  96, 361, 224, 489,  16, 281, 144, 409,
     80, 345, 208, 473,  48, 313, 176, 441, 112, 377,
    240, 505, 256,   8, 273, 136, 401,  72, 337, 200,
    465,  40, 305, 168, 433, 104, 369, 232, 497,  24,
    289, 152, 417,  88, 353, 216, 481,  56, 321, 184,
    449, 120, 385, 248, 513,   4, 269, 132, 397,  68,
    333, 196, 461,  36, 301, 164, 429, 100, 365, 228,
    493,  20, 285, 148, 413,  84, 349, 212, 477,  52,
    317, 180, 445, 116, 381, 244, 509, 261,  12, 277,
    140, 405,  76, 341, 204, 469,  44, 309, 172, 437,
    108, 373, 236, 501,  28, 293, 156, 421,  92, 357,
    220, 485,  60, 325, 188, 453, 124, 389, 252, 517,
      2, 267, 130, 395,  66, 331, 194, 459,  34, 299,
    162, 427,  98, 363, 226, 491,  18, 283, 146, 411,
     82, 347, 210, 475,  50, 315, 178, 443, 114, 379,
    242, 507, 258,  10, 275, 138, 403,  74, 339, 202,
    467,  42, 307, 170, 435, 106, 371, 234, 499,  26,
    291, 154, 419,  90, 355, 218, 483,  58, 323, 186,
    451, 122, 387, 250, 515,   6, 271, 134, 399,  70,
    335, 198, 463,  38, 303, 166, 431, 102, 367, 230,
    495,  22, 287, 150, 415,  86, 351, 214, 479,  54,
    319, 182, 447, 118, 383, 246, 511, 263,  14, 279,
    142, 407,  78, 343, 206, 471,  46, 311, 174, 439,
    110, 375, 238, 503,  30, 295, 158, 423,  94, 359,
    222, 487,  62, 327, 190, 455, 126, 391, 254, 519,
    260,   1, 266, 129, 394,  65, 330, 193, 458,  33,
    298, 161, 426,  97, 362, 225, 490,  17, 282, 145,
    410,  81, 346, 209, 474,  49, 314, 177, 442, 113,
    378, 241, 506, 257,   9, 274, 137, 402,  73, 338,
    201, 466,  41, 306, 169, 434, 105, 370, 233, 498,
     25, 290, 153, 418,  89, 354, 217, 482,  57, 322,
    185, 450, 121, 386, 249, 514,   5, 270, 133, 398,
     69, 334, 197, 462,  37, 302, 165, 430, 101, 366,
    229, 494,  21, 286, 149, 414,  85, 350, 213, 478,
     53, 318, 181, 446, 117, 382, 245, 510, 262,  13,
    278, 141, 406,  77, 342, 205, 470,  45, 310, 173,
    438, 109, 374, 237, 502,  29, 294, 157, 422,  93,
    358, 221, 486,  61, 326, 189, 454, 125, 390, 253,
    518,   3, 268, 131, 396,  67, 332, 195, 460,  35,
    300, 163, 428,  99, 364, 227, 492,  19, 284, 147,
    412,  83, 348, 211, 476,  51, 316, 179, 444, 115,
    380, 243, 508, 259,  11, 276, 139, 404,  75, 340,
    203, 468,  43, 308, 171, 436, 107, 372, 235, 500,
     27, 292, 155, 420,  91, 356, 219, 484,  59, 324,
    187, 452, 123, 388, 251, 516,   7, 272, 135, 400,
     71, 336, 199, 464,  39, 304, 167, 432, 103, 368,
    231, 496,  23, 288, 151, 416,  87, 352, 215, 480,
     55, 320, 183, 448, 119, 384, 247, 512, 264,  15,
    280, 143, 408,  79, 344, 207, 472,  47, 312, 175,
    440, 111, 376, 239, 504,  31, 296, 159, 424,  95,
    360, 223, 488,  63, 328, 191, 456, 127, 392, 255,
    520, 
    // Size 523
      0, 267, 128, 395,  64, 331, 192, 459,  32, 299,
    160, 427,  96, 363, 224, 491,  16, 283, 144, 411,
     80, 347, 208, 475,  48, 315, 176, 443, 112, 379,
    240, 507, 256,   8, 275, 136, 403,  72, 339, 200,
    467,  40, 307, 168, 435, 104, 371, 232, 499,  24,
    291, 152, 419,  88, 355, 216, 483,  56, 323, 184,
    451, 120, 387, 248, 515,   4, 271, 132, 399,  68,
    335, 196, 463,  36, 303, 164, 431, 100, 367, 228,
    495,  20, 287, 148, 415,  84, 351, 212, 479,  52,
    319, 180, 447, 116, 383, 244, 511, 263,  12, 279,
    140, 407,  76, 343, 204, 471,  44, 311, 172, 439,
    108, 375, 236, 503,  28, 295, 156, 423,  92, 359,
    220, 487,  60, 327, 188, 455, 124, 391, 252, 519,
    260,   2, 269, 130, 397,  66, 333, 194, 461,  34,
    301, 162, 429,  98, 365, 226, 493,  18, 285, 146,
    413,  82, 349, 210, 477,  50, 317, 178, 445, 114,
    381, 242, 509, 258,  10, 277, 138, 405,  74, 341,
    202, 469,  42, 309, 170, 437, 106, 373, 234, 501,
     26, 293, 154, 421,  90, 357, 218, 485,  58, 325,
    186, 453, 122, 389, 250, 517,   6, 273, 134, 401,
     70, 337, 198, 465,  38, 305, 166, 433, 102, 369,
    230, 497,  22, 289, 150, 417,  86, 353, 214, 481,
     54, 321, 182, 449, 118, 385, 246, 513, 265,  14,
    281, 142, 409,  78, 345, 206, 473,  46, 313, 174,
    441, 110, 377, 238, 505,  30, 297, 158, 425,  94,
    361, 222, 489,  62, 329, 190, 457, 126, 393, 254,
    521, 261,   1, 268, 129, 396,  65, 332, 193, 460,
     33, 300, 161, 428,  97, 364, 225, 492,  17, 284,
    145, 412,  81, 348, 209, 476,  49, 316, 177, 444,
    113, 380, 241, 508, 257,   9, 276, 137, 404,  73,
    340, 201, 468,  41, 308, 169, 436, 105, 372, 233,
    500,  25, 292, 153, 420,  89, 356, 217, 484,  57,
    324, 185, 452, 121, 388, 249, 516,   5, 272, 133,
    400,  69, 336, 197, 464,  37, 304, 165, 432, 101,
    368, 229, 496,  21, 288, 149, 416,  85, 352, 213,
    480,  53, 320, 181, 448, 117, 384, 245, 512, 264,
     13, 280, 141, 408,  77, 344, 205, 472,  45, 312,
    173, 440, 109, 376, 237, 504,  29, 296, 157, 424,
     93, 360, 221, 488,  61, 328, 189, 456, 125, 392,
    253, 520, 262,   3, 270, 131, 398,  67, 334, 195,
    462,  35, 302, 163, 430,  99, 366, 227, 494,  19,
    286, 147, 414,  83, 350, 211, 478,  51, 318, 179,
    446, 115, 382, 243, 510, 259,  11, 278, 139, 406,
     75, 342, 203, 470,  43, 310, 171, 438, 107, 374,
    235, 502,  27, 294, 155, 422,  91, 358, 219, 486,
     59, 326, 187, 454, 123, 390, 251, 518,   7, 274,
    135, 402,  71, 338, 199, 466,  39, 306, 167, 434,
    103, 370, 231, 498,  23, 290, 151, 418,  87, 354,
    215, 482,  55, 322, 183, 450, 119, 386, 247, 514,
    266,  15, 282, 143, 410,  79, 346, 207, 474,  47,
    314, 175, 442, 111, 378, 239, 506,  31, 298, 159,
    426,  95, 362, 223, 490,  63, 330, 191, 458, 127,
    394, 255, 522, 
    // Size 541
      0, 285, 128, 413,  64, 349, 192, 477,  32, 317,
    160, 445,  96, 381, 224, 509, 256,  16, 301, 144,
    429,  80, 365, 208, 493,  48, 333, 176, 461, 112,
    397, 240, 525, 264,   8, 293, 136, 421,  72, 357,
    200, 485,  40, 325, 168, 453, 104, 389, 232, 517,
    277,  24, 309, 152, 437,  88, 373, 216, 501,  56,
    341, 184, 469, 120, 405, 248, 533, 268,   4, 289,
    132, 417,  68, 353, 196, 481,  36, 321, 164, 449,
    100, 385, 228, 513, 260,  20, 305, 148, 433,  84,
    369, 212, 497,  52, 337, 180, 465, 116, 401, 244,
    529, 273,  12, 297, 140, 425,  76, 361, 204, 489,
     44, 329, 172, 457, 108, 393, 236, 521, 281,  28,
    313, 156, 441,  92, 377, 220, 505,  60, 345, 188,
    473, 124, 409, 252, 537,   2, 287, 130, 415,  66,
    351, 194, 479,  34, 319, 162, 447,  98, 383, 226,
    511, 258,  18, 303, 146, 431,  82, 367, 210, 495,
     50, 335, 178, 463, 114, 399, 242, 527, 266,  10,
    295, 138, 423,  74, 359, 202, 487,  42, 327, 170,
    455, 106, 391, 234, 519, 279,  26, 311, 154, 439,
     90, 375, 218, 503,  58, 343, 186, 471, 122, 407,
    250, 535, 271,   6, 291, 134, 419,  70, 355, 198,
    483,  38, 323, 166, 451, 102, 387, 230, 515, 262,
     22, 307, 150, 435,  86, 371, 214, 499,  54, 339,
    182, 467, 118, 403, 246, 531, 275,  14, 299, 142,
    427,  78, 363, 206, 491,  46, 331, 174, 459, 110,
    395, 238, 523, 283,  30, 315, 158, 443,  94, 379,
    222, 507,  62, 347, 190, 475, 126, 411, 254, 539,
    270,   1, 286, 129, 414,  65, 350, 193, 478,  33,
    318, 161, 446,  97, 382, 225, 510, 257,  17, 302,
    145, 430,  81, 366, 209, 494,  49, 334, 177, 462,
    113, 398, 241, 526, 265,   9, 294, 137, 422,  73,
    358, 201, 486,  41, 326, 169, 454, 105, 390, 233,
    518, 278,  25, 310, 153, 438,  89, 374, 217, 502,
     57, 342, 185, 470, 121, 406, 249, 534, 269,   5,
    290, 133, 418,  69, 354, 197, 482,  37, 322, 165,
    450, 101, 386, 229, 514, 261,  21, 306, 149, 434,
     85, 370, 213, 498,  53, 338, 181, 466, 117, 402,
    245, 530, 274,  13, 298, 141, 426,  77, 362, 205,
    490,  45, 330, 173, 458, 109, 394, 237, 522, 282,
     29, 314, 157, 442,  93, 378, 221, 506,  61, 346,
    189, 474, 125, 410, 253, 538,   3, 288, 131, 416,
     67, 352, 195, 480,  35, 320, 163, 448,  99, 384,
    227, 512, 259,  19, 304, 147, 432,  83, 368, 211,
    496,  51, 336, 179, 464, 115, 400, 243, 528, 267,
     11, 296, 139, 424,  75, 360, 203, 488,  43, 328,
    171, 456, 107, 392, 235, 520, 280,  27, 312, 155,
    440,  91, 376, 219, 504,  59, 344, 187, 472, 123,
    408, 251, 536, 272,   7, 292, 135, 420,  71, 356,
    199, 484,  39, 324, 167, 452, 103, 388, 231, 516,
    263,  23, 308, 151, 436,  87, 372, 215, 500,  55,
    340, 183, 468, 119, 404, 247, 532, 276,  15, 300,
    143, 428,  79, 364, 207, 492,  47, 332, 175, 460,
    111, 396, 239, 524, 284,  31, 316, 159, 444,  95,
    380, 223, 508,  63, 348, 191, 476, 127, 412, 255,
    540
};

const size_t* FaurePermutations[FaurePermutationTableSize] =
{
    FaurePermutationValues + 0,
    FaurePermutationValues + 2,
    FaurePermutationValues + 5,
    FaurePermutationValues + 10,
    FaurePermutationValues + 17,
    FaurePermutationValues + 28,
    FaurePermutationValues + 41,
    FaurePermutationValues + 58,
    FaurePermutationValues + 77,
    FaurePermutationValues + 100,
    FaurePermutationValues + 129,
    FaurePermutationValues + 160,
    FaurePermutationValues + 197,
    FaurePermutationValues + 238,
    FaurePermutationValues + 281,
    FaurePermutationValues + 328,
    FaurePermutationValues + 381,
    FaurePermutationValues + 440,
    FaurePermutationValues + 501,
    FaurePermutationValues + 568,
    FaurePermutationValues + 639,
    FaurePermutationValues + 712,
    FaurePermutationValues + 791,
    FaurePermutationValues + 874,
    FaurePermutationValues + 963,
    FaurePermutationValues + 1060,
    FaurePermutationValues + 1161,
    FaurePermutationValues + 1264,
    FaurePermutationValues + 1371,
    FaurePermutationValues + 1480,
    FaurePermutationValues + 1593,
    FaurePermutationValues + 1720,
    FaurePermutationValues + 1851,
    FaurePermutationValues + 1988,
    FaurePermutationValues + 2127,
    FaurePermutationValues + 2276,
    FaurePermutationValues + 2427,
    FaurePermutationValues + 2584,
    FaurePermutationValues + 2747,
    FaurePermutationValues + 2914,
    FaurePermutationValues + 3087,
    FaurePermutationValues + 3266,
    FaurePermutationValues + 3447,
    FaurePermutationValues + 3638,
    FaurePermutationValues + 3831,
    FaurePermutationValues + 4028,
    FaurePermutationValues + 4227,
    FaurePermutationValues + 4438,
    FaurePermutationValues + 4661,
    FaurePermutationValues + 4888,
    FaurePermutationValues + 5117,
    FaurePermutationValues + 5350,
    FaurePermutationValues + 5589,
    FaurePermutationValues + 5830,
    FaurePermutationValues + 6081,
    FaurePermutationValues + 6338,
    FaurePermutationValues + 6601,
    FaurePermutationValues + 6870,
    FaurePermutationValues + 7141,
    FaurePermutationValues + 7418,
    FaurePermutationValues + 7699,
    FaurePermutationValues + 7982,
    FaurePermutationValues + 8275,
    FaurePermutationValues + 8582,
    FaurePermutationValues + 8893,
    FaurePermutationValues + 9206,
    FaurePermutationValues + 9523,
    FaurePermutationValues + 9854,
    FaurePermutationValues + 10191,
    FaurePermutationValues + 10538,
    FaurePermutationValues + 10887,
    FaurePermutationValues + 11240,
    FaurePermutationValues + 11599,
    FaurePermutationValues + 11966,
    FaurePermutationValues + 12339,
    FaurePermutationValues + 12718,
    FaurePermutationValues + 13101,
    FaurePermutationValues + 13490,
    FaurePermutationValues + 13887,
    FaurePermutationValues + 14288,
    FaurePermutationValues + 14697,
    FaurePermutationValues + 15116,
    FaurePermutationValues + 15537,
    FaurePermutationValues + 15968,
    FaurePermutationValues + 16401,
    FaurePermutationValues + 16840,
    FaurePermutationValues + 17283,
    FaurePermutationValues + 17732,
    FaurePermutationValues + 18189,
    FaurePermutationValues + 18650,
    FaurePermutationValues + 19113,
    FaurePermutationValues + 19580,
    FaurePermutationValues + 20059,
    FaurePermutationValues + 20546,
    FaurePermutationValues + 21037,
    FaurePermutationValues + 21536,
    FaurePermutationValues + 22039,
    FaurePermutationValues + 22548,
    FaurePermutationValues + 23069,
    FaurePermutationValues + 23592
};

const PrimePermutationPair PrimesFaurePermutations[FaurePermutationTableSize] =
{
    { Primes[ 0], FaurePermutationValues + 0 },
    { Primes[ 1], FaurePermutationValues + 2 },
    { Primes[ 2], FaurePermutationValues + 5 },
    { Primes[ 3], FaurePermutationValues + 10 },
    { Primes[ 4], FaurePermutationValues + 17 },
    { Primes[ 5], FaurePermutationValues + 28 },
    { Primes[ 6], FaurePermutationValues + 41 },
    { Primes[ 7], FaurePermutationValues + 58 },
    { Primes[ 8], FaurePermutationValues + 77 },
    { Primes[ 9], FaurePermutationValues + 100 },
    { Primes[10], FaurePermutationValues + 129 },
    { Primes[11], FaurePermutationValues + 160 },
    { Primes[12], FaurePermutationValues + 197 },
    { Primes[13], FaurePermutationValues + 238 },
    { Primes[14], FaurePermutationValues + 281 },
    { Primes[15], FaurePermutationValues + 328 },
    { Primes[16], FaurePermutationValues + 381 },
    { Primes[17], FaurePermutationValues + 440 },
    { Primes[18], FaurePermutationValues + 501 },
    { Primes[19], FaurePermutationValues + 568 },
    { Primes[20], FaurePermutationValues + 639 },
    { Primes[21], FaurePermutationValues + 712 },
    { Primes[22], FaurePermutationValues + 791 },
    { Primes[23], FaurePermutationValues + 874 },
    { Primes[24], FaurePermutationValues + 963 },
    { Primes[25], FaurePermutationValues + 1060 },
    { Primes[26], FaurePermutationValues + 1161 },
    { Primes[27], FaurePermutationValues + 1264 },
    { Primes[28], FaurePermutationValues + 1371 },
    { Primes[29], FaurePermutationValues + 1480 },
    { Primes[30], FaurePermutationValues + 1593 },
    { Primes[31], FaurePermutationValues + 1720 },
    { Primes[32], FaurePermutationValues + 1851 },
    { Primes[33], FaurePermutationValues + 1988 },
    { Primes[34], FaurePermutationValues + 2127 },
    { Primes[35], FaurePermutationValues + 2276 },
    { Primes[36], FaurePermutationValues + 2427 },
    { Primes[37], FaurePermutationValues + 2584 },
    { Primes[38], FaurePermutationValues + 2747 },
    { Primes[39], FaurePermutationValues + 2914 },
    { Primes[40], FaurePermutationValues + 3087 },
    { Primes[41], FaurePermutationValues + 3266 },
    { Primes[42], FaurePermutationValues + 3447 },
    { Primes[43], FaurePermutationValues + 3638 },
    { Primes[44], FaurePermutationValues + 3831 },
    { Primes[45], FaurePermutationValues + 4028 },
    { Primes[46], FaurePermutationValues + 4227 },
    { Primes[47], FaurePermutationValues + 4438 },
    { Primes[48], FaurePermutationValues + 4661 },
    { Primes[49], FaurePermutationValues + 4888 },
    { Primes[50], FaurePermutationValues + 5117 },
    { Primes[51], FaurePermutationValues + 5350 },
    { Primes[52], FaurePermutationValues + 5589 },
    { Primes[53], FaurePermutationValues + 5830 },
    { Primes[54], FaurePermutationValues + 6081 },
    { Primes[55], FaurePermutationValues + 6338 },
    { Primes[56], FaurePermutationValues + 6601 },
    { Primes[57], FaurePermutationValues + 6870 },
    { Primes[58], FaurePermutationValues + 7141 },
    { Primes[59], FaurePermutationValues + 7418 },
    { Primes[60], FaurePermutationValues + 7699 },
    { Primes[61], FaurePermutationValues + 7982 },
    { Primes[62], FaurePermutationValues + 8275 },
    { Primes[63], FaurePermutationValues + 8582 },
    { Primes[64], FaurePermutationValues + 8893 },
    { Primes[65], FaurePermutationValues + 9206 },
    { Primes[66], FaurePermutationValues + 9523 },
    { Primes[67], FaurePermutationValues + 9854 },
    { Primes[68], FaurePermutationValues + 10191 },
    { Primes[69], FaurePermutationValues + 10538 },
    { Primes[70], FaurePermutationValues + 10887 },
    { Primes[71], FaurePermutationValues + 11240 },
    { Primes[72], FaurePermutationValues + 11599 },
    { Primes[73], FaurePermutationValues + 11966 },
    { Primes[74], FaurePermutationValues + 12339 },
    { Primes[75], FaurePermutationValues + 12718 },
    { Primes[76], FaurePermutationValues + 13101 },
    { Primes[77], FaurePermutationValues + 13490 },
    { Primes[78], FaurePermutationValues + 13887 },
    { Primes[79], FaurePermutationValues + 14288 },
    { Primes[80], FaurePermutationValues + 14697 },
    { Primes[81], FaurePermutationValues + 15116 },
    { Primes[82], FaurePermutationValues + 15537 },
    { Primes[83], FaurePermutationValues + 15968 },
    { Primes[84], FaurePermutationValues + 16401 },
    { Primes[85], FaurePermutationValues + 16840 },
    { Primes[86], FaurePermutationValues + 17283 },
    { Primes[87], FaurePermutationValues + 17732 },
    { Primes[88], FaurePermutationValues + 18189 },
    { Primes[89], FaurePermutationValues + 18650 },
    { Primes[90], FaurePermutationValues + 19113 },
    { Primes[91], FaurePermutationValues + 19580 },
    { Primes[92], FaurePermutationValues + 20059 },
    { Primes[93], FaurePermutationValues + 20546 },
    { Primes[94], FaurePermutationValues + 21037 },
    { Primes[95], FaurePermutationValues + 21536 },
    { Primes[96], FaurePermutationValues + 22039 },
    { Primes[97], FaurePermutationValues + 22548 },
    { Primes[98], FaurePermutationValues + 23069 },
    { Primes[99], FaurePermutationValues + 23592 }
};

}   // namespace foundation
